/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.Execution;



/**
 * Implementation base of a body forwarder. This class handles closing of he data source and exceptions by 
 * forwarding the body data  
 * 
 * @author grro@xsocket.org
 */
public abstract class AbstractBodyForwarder implements IBodyDataHandler {
	
	private static final Logger LOG = Logger.getLogger(AbstractBodyForwarder.class.getName());
	
	public static final int DEFAULT_EXECUTION_MODE = IBodyDataHandler.DEFAULT_EXECUTION_MODE;
	
	
	private final ForwardTask forwardTask = new ForwardTask();
	
	private NonBlockingBodyDataSource bodyDataSource = null;
	private BodyDataSink bodyDataSink = null;
	
	public AbstractBodyForwarder(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) {
		this.bodyDataSource = bodyDataSource;
		this.bodyDataSink = bodyDataSink;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Execution(Execution.NONTHREADED)
	public final boolean onData(final NonBlockingBodyDataSource bodyDataSource) throws BufferUnderflowException {
		
		assert (this.bodyDataSource == bodyDataSource);
		
		try {
			int available = bodyDataSource.available();
			
			// data available
			if (available > 0) {
				if (HttpUtils.isMutlithreaded(this)) {
					bodyDataSource.processMultiThreaded(forwardTask);
				} else {
					bodyDataSource.processNonThreaded(forwardTask);
				}
				
			// end of stream reached
			} else if (available == -1) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("bodyDataSource is closed (available == -1). calling on complete and closing bodyDataSink"); 
				}

				callOnComplete();
				closeDataSink();
			}

		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + bodyDataSource.getId() + "] error occured by writing data sink " + bodyDataSource + " " + e.toString() + " destroying data source and data sink");
			}
			handleException(e);
		}
		
		return true;
	}

	private void closeDataSink() {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + bodyDataSource.getId() + "] data source " + bodyDataSource.getClass().getSimpleName() + " is closed. Closing data sink " + bodyDataSink.getClass().getSimpleName());
		}
		try {
			bodyDataSink.close();
		} catch (Exception e) {
			handleException(e);
		}
	}
	
	private void handleException(Exception e) {
		bodyDataSource.destroy();
		bodyDataSink.destroy();
		
		onException(e);  
	}
	
	
	
	/**
	 * call back method which will be called if the body data source contains a least one byte
	 * 
	 * @param bodyDataSource   the body data source 
	 * @param bodyDataSink     the body data sink
	 * @throws BufferUnderflowException if not enough data is available 
	 * @throws IOException if an exception occurs 
	 */
	public void onData(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) throws BufferUnderflowException, IOException {
		ByteBuffer[] data = bodyDataSource.readByteBufferByLength(bodyDataSource.available());
		
		bodyDataSink.write(data);
		bodyDataSink.flush();
	}
	
	
	private void callOnComplete() {
		try {
			onComplete();
		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by calling onComplete "  + e.toString());
				throw new RuntimeException(e.toString());
			}
		}
	}
		
	public void onComplete() {
		
	}

	public void onException(Exception e) {
		
	}

	
	private final class ForwardTask implements Runnable {
			
		public void run() {
			try {
				while (bodyDataSource.available() > 0) {
					int version = bodyDataSource.getReadBufferVersion();
					
					onData(bodyDataSource, bodyDataSink);
						
					int newVersion = bodyDataSource.getReadBufferVersion();
					if (newVersion != version) {
						version = newVersion;
						
					} else {
						break;
					}
				}
				
				if (bodyDataSource.available() == -1) {
					callOnComplete();
					closeDataSink();
				}
				
			} catch (Exception e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by forwarding data " + DataConverter.toString(e));
				}
				handleException(e);
			}
		}		
	}
}
