/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.net.ConnectException;
import java.net.InetAddress;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.logging.Level;
import java.util.logging.Logger;


import org.xsocket.DataConverter;
import org.xsocket.Execution;
import org.xsocket.MaxReadSizeExceededException;
import org.xsocket.SerializedTaskQueue;
import org.xsocket.connection.IConnectHandler;
import org.xsocket.connection.IConnectionTimeoutHandler;
import org.xsocket.connection.IDataHandler;
import org.xsocket.connection.IDisconnectHandler;
import org.xsocket.connection.IHandler;
import org.xsocket.connection.IIdleTimeoutHandler;
import org.xsocket.connection.INonBlockingConnection;
import org.xsocket.connection.NonBlockingConnectionPool;
import org.xsocket.connection.http.AbstractHttpMessage.BodyType;
import org.xsocket.connection.http.HttpUtils.HttpHandlerInfo;
import org.xsocket.connection.http.client.HttpClient;
import org.xsocket.connection.http.client.HttpClientConnection;
import org.xsocket.connection.http.server.HttpServerConnection;





/**
 * 
 * Implementation base for the {@link HttpClientConnection} and {@link HttpServerConnection}
 * 
 * <br/><br/><b>This is a xSocket internal class and subject to change</b> 
 * 
 * @author grro@xsocket.org
 */
public abstract class AbstractHttpConnection implements IHttpConnection {

	private static final Logger LOG = Logger.getLogger(AbstractHttpConnection.class.getName());
		
	protected static final int MAX_HEADER_LENGTH = Integer.MAX_VALUE;

	public static final String DEFAULT_BODY_ENCODING = AbstractMessageHeader.DEFAULT_BODY_ENCODING;
	
	private static Executor defaultWorkerPool = null;

	
	// flags
	private boolean isOpen = true;
	private final AtomicBoolean isSuspended = new AtomicBoolean(false);
	
	
	// timer support 
	private static final Timer TIMER = new Timer("xHttpTimer", true);

		
	// underlying connection
	private INonBlockingConnection tcpConnection = null;
	
	
	// protocol handler
	private final ProtocolHandler protocolHandler = new ProtocolHandler();
	
		
	// flags
	private boolean isPersistent = false;
	
	
	
	// attached parser
	private IDataHandler bodyParser = null;
	
		
	// output body close listener support
	private final AtomicReference<BodyDataSink> bodyDataSink = new AtomicReference<BodyDataSink>();
	
	
	// attachment
	private Object attachment = null;
	
	
	
	// call back processing
	private final SerializedTaskQueue taskQueue = new SerializedTaskQueue();
	
	private long lastTimeDataWritten = System.currentTimeMillis();
	

	/**
	 * constructor 
	 * 
	 * @param nativeConnection   the underlying tcp connection
	 * @throws IOException if an exception occurs
	 */
	protected AbstractHttpConnection(INonBlockingConnection nativeConnection) throws IOException {
		this.tcpConnection = nativeConnection;
		
		tcpConnection.setAutoflush(false);
		tcpConnection.setAttachment(this);
	}
	
	protected final void init() throws IOException {
		protocolHandler.onConnect(tcpConnection);
		tcpConnection.setHandler(protocolHandler);		
	}


	/**
	 * retrieve a http connection based on a given native conection (http connection will be attached 
	 * to the native connection within the http connection constructor)
	 * 
	 * @param nativeConnection  the native connection
	 * @return the http connection
	 */
	static AbstractHttpConnection getHttpConnection(INonBlockingConnection nativeConnection) {
		return (AbstractHttpConnection) nativeConnection.getAttachment();
	}
	
	protected final static boolean isComplete(NonBlockingBodyDataSource body) throws IOException {
		return body.isComplete();
	}
	
	
	static boolean hasChunkedBody(AbstractHttpMessage message) throws IOException {
		return message.getBodyType() == BodyType.CHUNKED;
	}

	static boolean hasBoundBody(AbstractHttpMessage message) throws IOException {
		return message.getBodyType() == BodyType.BOUND;
	}
	
	protected final static boolean hasConnectionTerminatedBody(HttpResponse response) throws IOException {
		return (response.getBodyType() == BodyType.CONNECTION_TERMINATED);
	}

	protected final BodyDataSink newBoundBodyDataSink(IHttpHeader header) throws IOException {
		return new BodyDataSink(this, header, new BoundBodyWriter(this, header));
	}
	
	protected final BodyDataSink newChunkedBodyDataSink(IHttpHeader header) throws IOException {
		return new BodyDataSink(this, header, new ChunkedBodyWriter(this, header));
	}
	
	protected static final HttpResponseHeader newEmptyResponseHeader() {
		return new HttpResponseHeader();
	}
	
	
	protected static final void newClientExhange(HttpClient httpClient, HttpClientConnection con, IHttpRequest request, IHttpRequestHandler requestHandler, boolean isRequestHandlerChain, IHttpResponseHandler responseHandler) throws IOException {
		new ClientExchange(httpClient, con, request, requestHandler, isRequestHandlerChain, responseHandler).handle();
	}
	
	protected final IHttpRequest newFormEncodedRequestWrapper(IHttpRequest request) throws IOException {
		return HttpUtils.newFormEncodedRequestWrapper(request);
	}
	
	protected final BodyDataSink newClientInterception(HttpClient httpClient, HttpClientConnection con, IHttpRequestHeader requestHeader, IHttpRequestHandler interceptor, boolean isInterceptorChain, IHttpResponseHandler responseHandler) throws IOException {

		final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(requestHeader.getCharacterEncoding());
		HttpRequest request = new HttpRequest(requestHeader, bodyDataSource);
		
		IBodyWriter bodyWriter = new IBodyWriter() {
			 
			public void flush(ByteBuffer[] bodyData) throws IOException {
				bodyDataSource.append(bodyData);
			}
			
			public void close() throws IOException {
				bodyDataSource.setComplete(true);
			}
			
			public void destroy() {
				bodyDataSource.destroy();
			}
		};
		
		BodyDataSink bodyDataSink = new BodyDataSink(null, requestHeader, bodyWriter);
		
		new ClientExchange(httpClient, con, request, interceptor, isInterceptorChain, responseHandler).handle();
		
		return bodyDataSink;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public int getMaxReadBufferThreshold() {
		return tcpConnection.getMaxReadBufferThreshold();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setMaxReadBufferThreshold(int size) {
		tcpConnection.setMaxReadBufferThreshold(size);
	}
	
	
	
	/**
	 * returns the default worker pool
	 * 
	 * @return the default worker pool
	 */
	synchronized static Executor getDefaultWorkerpool() {
		if (defaultWorkerPool == null) {
			defaultWorkerPool = Executors.newCachedThreadPool(new DefaultThreadFactory());
		}
		
		return defaultWorkerPool;
	}

	
	protected  static void schedule(TimerTask task, long delay, long period) {
		TIMER.schedule(task, delay, period);
	}
	
	
	public final Executor getWorkerpool() {
		return tcpConnection.getWorkerpool();
	}

	protected  void suspendRead() throws IOException {
		isSuspended.set(true);
		tcpConnection.suspendRead();
	}

	protected void resumeRead() throws IOException {
		isSuspended.set(false);
		tcpConnection.resumeRead();
	}
	
	
	
	boolean isReadSuspended() {
		return isSuspended.get();
	}

	
	protected  final long getLastTimeWritten() {
		return lastTimeDataWritten;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final boolean isSecure() {
		return tcpConnection.isSecure();
	}
	
	
	/**
	 * returns if the connection is persistent 
	 * 
	 * @return true, if the connection is persistent
	 */
	public final boolean isPersistent() {
		return isPersistent;
	}
	
	
	/**
	 * sets the persistent flag
	 * 
	 * @param isPersistent the persistent flag 
	 */
	protected final void setPersistent(boolean isPersistent) {
		this.isPersistent = isPersistent;
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setOption(String name, Object value) throws IOException {
		tcpConnection.setOption(name, value);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final Object getOption(String name) throws IOException {
		return tcpConnection.getOption(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final Map<String, Class> getOptions() {
		return tcpConnection.getOptions();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public final void setConnectionTimeoutMillis(long timeoutMillis) {
		tcpConnection.setConnectionTimeoutMillis(timeoutMillis);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final long getConnectionTimeoutMillis() {
		return tcpConnection.getConnectionTimeoutMillis();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setIdleTimeoutMillis(long timeoutMillis) {
		tcpConnection.setIdleTimeoutMillis(timeoutMillis);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final long getIdleTimeoutMillis() {
		return tcpConnection.getIdleTimeoutMillis();
	}
	
	
	final String dumpReadChannel() {
		try {
			return tcpConnection.readStringByLength(tcpConnection.available());
		} catch (IOException ioe) {
			return "error occured by dumping read channel";
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final boolean isOpen() {
		if (!isOpen) {
			return false;
		}
		
		return tcpConnection.isOpen();
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	public final String getId() {
		return tcpConnection.getId();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public long getRemainingMillisToConnectionTimeout() {
		return tcpConnection.getRemainingMillisToConnectionTimeout();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public long getRemainingMillisToIdleTimeout() {
		return tcpConnection.getRemainingMillisToIdleTimeout();
	}
	
	protected final void processMultiThreaded(Runnable task) {
		taskQueue.performMultiThreaded(task, tcpConnection.getWorkerpool());
	}
	
	protected final void processNonThreaded(Runnable task) {
		taskQueue.performNonThreaded(task);
	}
	
	
	/**
	 * ad hoc activation of a secured mode (SSL). By performing of this
	 * method all remaining data to send will be flushed.
	 * After this all data will be sent and received in the secured mode
	 *
	 * @throws IOException If some other I/O error occurs
	 */
	public final void activateSecuredMode() throws IOException {
		tcpConnection.activateSecuredMode();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void close() throws IOException {

		Runnable closeTask = new Runnable() {
			
			public void run() {
				doClose();	
			}		
		};

		processNonThreaded(closeTask);
	}
	
	
	private void doClose() {
		try {
			if (isResusable()) {
				tcpConnection.close();
				
			} else {
				
				if (LOG.isLoggable(Level.FINE)) {
					if (getBodyParser() != null) {
						LOG.fine("destroying connection instead of closing it (BodyParser is set)");
					}
				}
				
				// destroy (underlying tcp) connection by using connection pool. The underlying connection could be a pooled one)
				// The connection pool detects automatically if the connection is pooled or not. The connection will be 
				// closed (logically) anyway
				NonBlockingConnectionPool.destroy(tcpConnection);
			}
		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by closing htttp connection " + getId() + " " + e.toString());
			}
		}			
	}
	
	
	boolean isResusable() {
		return (getBodyParser() == null) && isPersistent; 
	}
	

	protected  final void closeSilence() {
		try {
			close();
		} catch (IOException ioe) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by closing connection " + getId() + " " + ioe.toString());
			}
			
			try {
				NonBlockingConnectionPool.destroy(tcpConnection);
			} catch (IOException ignore) { }
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setAttachment(Object obj) {
		this.attachment = obj;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final Object getAttachment() {
		return attachment;
	}
	
	 

	public final void setWriteTransferRate(int bytesPerSecond) throws ClosedChannelException, IOException {
		tcpConnection.setWriteTransferRate(bytesPerSecond);
	}
	
	
	
	
	public final int getWriteTransferRate() throws ClosedChannelException, IOException {
		return tcpConnection.getWriteTransferRate();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public final InetAddress getLocalAddress() {
		return tcpConnection.getLocalAddress();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final int getLocalPort() {
		return tcpConnection.getLocalPort();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final InetAddress getRemoteAddress() {
		return tcpConnection.getRemoteAddress();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final int getRemotePort() {
		return tcpConnection.getRemotePort();
	}
	
	

	/**
	 * destroy the connection
	 */
	public void destroy() {
		
		isPersistent = false;
		
		Runnable destroyTask = new Runnable() {
			
			public void run() {
				// destroy (underlying tcp) connection by using connection pool. The underlying connection could be a pooled one)
				// The connection pool detects automatically if the connection is pooled or not. The connection will be 
				// closed (logically) anyway
				try {
					NonBlockingConnectionPool.destroy(tcpConnection);
				} catch (IOException ioe) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("error occured by destroying htttp connection " + getId() + " " + ioe.toString());
					}
				}				
			}		
		};

		processNonThreaded(destroyTask);
	}
	

	
	public final void setFlushmode(FlushMode flushmode) {
		tcpConnection.setFlushmode(flushmode);
	}
	
	public final FlushMode getFlushmode() {
		return tcpConnection.getFlushmode();
	}
	
	protected final void flush() throws IOException {
		tcpConnection.flush();
	}
	
	final int write(String txt) throws IOException {
		lastTimeDataWritten = System.currentTimeMillis();
		return tcpConnection.write(txt);
	}
	
	final long write(ByteBuffer[] buffer) throws IOException {
		lastTimeDataWritten = System.currentTimeMillis();
		return tcpConnection.write(buffer);
	}
	
	protected final int writeHeader(IHttpHeader header) throws IOException {
		lastTimeDataWritten = System.currentTimeMillis();
		
		int written = header.writeTo(tcpConnection);
		return written;
	}
	
	
	protected  final void sendMessageBody(BodyDataSink bodyDataSink, NonBlockingBodyDataSource bodyDataSource) throws IOException {
		
		try {
			bodyDataSink.setFlushmode(FlushMode.ASYNC);
			bodyDataSink.setAutoflush(false);
	
			// is body complete?
			if (bodyDataSource.isComplete()) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("message body to sent is complete.writing all data to body data sink ");
				}
				
				int available = bodyDataSource.available();
				if (available > 0) {
					bodyDataSink.write(bodyDataSource.readBytesByLength(available));
				}
				bodyDataSink.close();	
	
			// .. no
			} else {
				BodyDataForwardHandler forwardHandler = new BodyDataForwardHandler(bodyDataSource, bodyDataSink);
				bodyDataSource.setDataHandler(forwardHandler);
				bodyDataSink.flush();
			}
			
		} catch (ClosedChannelException cce) {
			throw new IOException("http connection " + getId() + " is already closed " + cce.toString());
		}
	}


	
	protected void onBodyDataReceived() {
		
	}
	
	
	void setBodyDataSink(BodyDataSink bodyChannel) {
		bodyDataSink.set(bodyChannel);
	}
	
	
	boolean removeBodyDataSink(BodyDataSink bodyChannel) {
		BodyDataSink bds = bodyDataSink.get(); 
		if ((bds != null) && (bds == bodyChannel)) {
			bodyDataSink.set(null);
			return true;
		}
		
		return false;
	}
	

/*
	protected final BodyDataSink replaceBody(HttpRequest httpRequest) throws IOException {
		
		// create a new data source object
		final NonBlockingBodyDataSource dataSource = new NonBlockingBodyDataSource(httpRequest.getCharacterEncoding());
		httpRequest.setBodyDataSource(dataSource);
		

		IBodyWriter bodyToDataSourceWriter = new IBodyWriter() {
			
			public void flush(ByteBuffer[] bodyData) {
				dataSource.append(bodyData);
			}
			
			public void close() {
				dataSource.setComplete(true);		
			}
			
			public void destroy() {
				dataSource.destroy();
			}
		};
		
		
		return new BodyDataSink(this, httpRequest.getRequestHeader(), bodyToDataSourceWriter);
	}
	*/

/*	
	protected final void setInvokeOnMessage(IHttpResponse response) {
		
		if (response instanceof HttpResponseView) {

			HttpResponseView rv = (HttpResponseView) response;
			
			final IBodyWriter bodyWriter = rv.getBodyWriter();
			
			IBodyWriter onCompelteBodyWriter = new IBodyWriter() {
				
				private final List<ByteBuffer> buffers = new ArrayList<ByteBuffer>();
				
				public void flush(ByteBuffer[] bodyData) {
					for (ByteBuffer byteBuffer : bodyData) {
						buffers.add(byteBuffer);
					}
				}
				
				public void close() throws IOException {
					bodyWriter.close();		
				}
				
				public void destroy() {
					bodyWriter.destroy();
				}
			};
			rv.setBodyWriter(onCompelteBodyWriter);
			
		}
	}
	*/
	
	/*
	protected final BodyDataSink replaceBody(IHttpResponse httpResponse) throws IOException {
		
		if (httpResponse instanceof HttpResponseView) {
			return ((HttpResponseView) httpResponse).replaceBody(this);
			
		} else {
			
			return null;
		}
	}
	*/
	
	
	protected abstract void onMessage(INonBlockingConnection connection) throws BufferUnderflowException, IOException;
	
	

	

	protected final void addBoundBodyParser(AbstractHttpMessage message) throws IOException {
		NonBlockingBodyDataSource dataSource = new NonBlockingBodyDataSource(message.getCharacterEncoding(), this);
		message.setBodyDataSource(dataSource);
		
		BoundBodyParser parser = new BoundBodyParser(this, message.getMessageHeader(), dataSource);
		parser.onData(tcpConnection);
	}
	
	protected final void addChunkedBodyParser(AbstractHttpMessage message) throws IOException {
		NonBlockingBodyDataSource dataSource = new NonBlockingBodyDataSource(message.getCharacterEncoding(), this);
		message.setBodyDataSource(dataSource);
		
		ChunkedBodyParser parser = new ChunkedBodyParser(this, message.getMessageHeader(), dataSource);
		parser.onData(tcpConnection);
	}
	
	protected final void addConnectionTerminatedBodyParser(AbstractHttpMessage message, INonBlockingConnection connection) throws IOException {
		setPersistent(false);
		
		NonBlockingBodyDataSource dataSource = new NonBlockingBodyDataSource(message.getCharacterEncoding(), this);
		message.setBodyDataSource(dataSource);
		
		ConnectionTerminatedBodyParser parser = new ConnectionTerminatedBodyParser(this, dataSource);
		parser.onData(connection);
	}
	
	

	/**
	 * disconnect call back 
	 * 
	 * @throws IOException  if an exception occurs
	 */
	protected void onDisconnect() throws IOException {

		BodyDataSink bds = bodyDataSink.get();
		if (bds != null) {
			bds.onUnderlyingHttpConnectionClosed();
			bodyDataSink.set(null);
		}		
	}
	
	
	/**
	 * idle time out call back 
	 * 
	 * @throws IOException  if an exception occurs
	 */
	private void onIdleTimeout() throws IOException {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId()  + "] idle timeout reached (" + 
					 DataConverter.toFormatedDuration(tcpConnection.getIdleTimeoutMillis()) + 
					 "). terminate connection");
		}

		this.close();
	}
	
	
	/**
	 * connection timeout call back 
	 * 
	 * @throws IOException  if an exception occurs
	 */
	protected void onConnectionTimeout() throws IOException {
		this.close();
	}
	
	
	protected void onProtocolException(Throwable throwable) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId() + "] protocol error. closing connection " + throwable.toString());
		}
		
		
		destroy();
	}



		
	
	/**
	 * remove the current body parser
	 */
	void removeBodyParser() {
		bodyParser = null;
	}

	
	private IHandler getBodyParser() {
		return bodyParser;
	}


	
	void setBodyParser(IDataHandler parser) {
		this.bodyParser = parser;
	}
	
	

	protected  static String getReason(int statusCode) {
		switch (statusCode) {
		case 200:
			return "OK";

		case 401:
			return "Unauthorized";

		case 404:
			return "Not found";
			
		case 500:
			return "Internal Server Error";
			
		case 501:
			return "Not Implemented";

		case 502:
			return "Bad Gateway";

		case 503:
			return "Service Unavailable";
			
		case 504:
			return "Gateway Timeout";
		
		case 505:
			return "HTTP Version Not Supported";

			
		default:
			return " ";
		}
	}
	


	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {		
		return tcpConnection.toString();
	}
	
	

	
	
	@Execution(Execution.NONTHREADED)
	private final class ProtocolHandler implements IDataHandler, IConnectHandler, IDisconnectHandler, IIdleTimeoutHandler, IConnectionTimeoutHandler {

		
		public boolean onData(INonBlockingConnection connection) throws BufferUnderflowException {

			try {
				if (connection.available() <= 0) {
					return true;
					
				}
			
				AbstractHttpConnection httpCon = getHttpConnection(connection);
				if (httpCon != null) {
						
					IDataHandler bodyParser = (IDataHandler) httpCon.getBodyParser();
						
					// body handling?
					if (bodyParser != null) {
						onBodyDataReceived();
						return bodyParser.onData(connection);
						
					} 
						
					AbstractHttpConnection.this.onMessage(connection);
						
				} else {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("no http connection assigned. closing tcp connection");
					}

					connection.close();
				}
				
			} catch (ClosedChannelException ignore) {
				
				
			} catch (BufferUnderflowException bue) {
				throw bue;				
				
			} catch (Exception ex) {
				onProtocolException(ex);
			}
			
			return true;
		}
		
		public boolean onConnect(INonBlockingConnection connection) throws IOException, BufferUnderflowException, MaxReadSizeExceededException {
			return true;
		}
		
		public boolean onDisconnect(INonBlockingConnection connection) throws IOException {
			isOpen = false;
			
			IDisconnectHandler bodyParser = (IDisconnectHandler) getHttpConnection(connection).getBodyParser();

			// body handling?
			if (bodyParser != null) {
				bodyParser.onDisconnect(connection);
			}
			
			AbstractHttpConnection.this.onDisconnect();
			return true;
		}
		
		
		public boolean onConnectionTimeout(INonBlockingConnection connection) throws IOException {
			AbstractHttpConnection.this.onConnectionTimeout();
			return true;
		}
		
		
		public boolean onIdleTimeout(INonBlockingConnection connection) throws IOException {
			AbstractHttpConnection.this.onIdleTimeout();
			return true;
		}
	}
	
	
	private final class BodyDataForwardHandler extends AbstractBodyForwarder {

		private int forwarded = 0;
		
		
		public BodyDataForwardHandler(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) {
			super(bodyDataSource, bodyDataSink);
			bodyDataSink.setFlushmode(FlushMode.ASYNC);
		}
		
		
		@Execution(Execution.NONTHREADED)
		public void onData(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) throws BufferUnderflowException, IOException {
			long written = bodyDataSink.write(bodyDataSource.readByteBufferByLength(bodyDataSource.available()));
			bodyDataSink.flush();
					
			forwarded += written;
					
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getId() + "] " + bodyDataSource.getClass().getSimpleName() + " " + written + " bytes forwarded to data sink " + bodyDataSink.getClass().getSimpleName() + " (total forwarded=" + forwarded + ")");
			}
		}
	};
	
	
	private static class DefaultThreadFactory implements ThreadFactory {
		private static final AtomicInteger poolNumber = new AtomicInteger(1);
		private final ThreadGroup group;
		private final AtomicInteger threadNumber = new AtomicInteger(1);
		private final String namePrefix;

		DefaultThreadFactory() {
			SecurityManager s = System.getSecurityManager();
			group = (s != null)? s.getThreadGroup() : Thread.currentThread().getThreadGroup();
			namePrefix = "xHttpNbcPool-" + poolNumber.getAndIncrement() + "-thread-";
        }

		public Thread newThread(Runnable r) {
			Thread t = new Thread(group, r, namePrefix + threadNumber.getAndIncrement(), 0);
            if (!t.isDaemon()) {
                t.setDaemon(true);
            }
            if (t.getPriority() != Thread.NORM_PRIORITY) {
                t.setPriority(Thread.NORM_PRIORITY);
            }
            return t;
        }
    };
    
    

    private static final class ClientExchange implements IHttpExchange {

		private static final Logger LOG = Logger.getLogger(ClientExchange.class.getName());
		
		
		private boolean isResponseCommitted = false; 
		
		
		
		private HttpClientConnection con = null;
		private IHttpRequest request = null;
		private String targetURL = null;
		private IHttpRequestHandler requestHandler = null;
		private HttpHandlerInfo requestHandlerInfo = null;
		private IHttpResponseHandler responseHandler = null;
		private HttpHandlerInfo responseHandlerInfo = null;
		private HttpClient httpClient = null;
		
		public ClientExchange(HttpClient httpClient, final HttpClientConnection con, IHttpRequest request, IHttpRequestHandler requestHandler, boolean isRequestHandlerChain, IHttpResponseHandler responseHandler) {
			this.httpClient = httpClient; 
			this.con = con;
			this.request = request;
			this.requestHandler = requestHandler;
			this.responseHandler = responseHandler;
			
			targetURL = request.getTargetURL().toString();
			if (targetURL.indexOf("?") != -1) {
				targetURL = targetURL.substring(0, targetURL.indexOf("?"));
			}
			
			requestHandlerInfo = HttpUtils.getHttpHandlerInfo(requestHandler);
			
			if (responseHandler != null) {
				responseHandlerInfo = HttpUtils.getHttpHandlerInfo(responseHandler);
			}
			
			
			if (isRequestHandlerChain) {
				IHttpRequestHandler reqHdl = new IHttpRequestHandler() {
					
					public void onRequest(final IHttpExchange exchange) throws IOException {
						
						IHttpResponseHandler respHdl = new IHttpResponseHandler() {
							
							public void onResponse(IHttpResponse response) throws IOException {
								exchange.send(response);
							}
							
							public void onException(IOException ioe) {
								exchange.sendError(500);
							}
						};
						
						con.send(exchange.getRequest(), respHdl);
					}
				};
				
				((RequestHandlerChain) requestHandler).addLast(reqHdl);

				
			}
			
		}
		

		void handle() throws IOException {
			
			final boolean isContentTypeFormUrlencoded = HttpUtils.isContentTypeFormUrlencoded(getRequest());
			
			if (requestHandlerInfo.isRequestHandlerInvokeOnMessageReceived() || isContentTypeFormUrlencoded) {
				
				IBodyCompleteListener cl = new IBodyCompleteListener() {
					
					public void onComplete() throws IOException {
						
						if (isContentTypeFormUrlencoded) {
							request = HttpUtils.newFormEncodedRequestWrapper(request);
						}

						callRequestHandler();
					}
				};
				request.getNonBlockingBody().addCompleteListener(cl);
				
			} else { 
				callRequestHandler();
			}
		}

		private void callRequestHandler() {

			
			Runnable task = new Runnable() {
				
				public void run() {
					try {
						requestHandler.onRequest(ClientExchange.this);
					} catch (IOException ioe) {
						if (LOG.isLoggable(Level.FINE)) {
							LOG.fine("error occured by calling request handler " + requestHandler + ". reason " + DataConverter.toString(ioe));
						}
					}					
				}
			};
			
			if (requestHandlerInfo.isRequestHandlerMultithreaded()) {
				con.processMultiThreaded(task);
			} else {
				con.processNonThreaded(task);
			}
		}
		
		
		
		public IHttpConnection getConnection() {
			return con;
		}

		public IHttpRequest getRequest() {
			return request;
		}

		
		public void destroy() {
			con.destroy();
		}

		
		public BodyDataSink send(IHttpResponseHeader header) throws IOException, IllegalStateException {
			
			if (responseHandler == null) {
				LOG.warning("response will not been send, because no response handler is assigned");
				return new BodyDataSink(null, header, null);
			}
			
			if (header.getContentLength() != -1) {
				LOG.warning("message header contains content length. removing it because message should be sent as chunked");
				header.removeHeader("Content-Length");
			}
			
			if ((header.getTransferEncoding() == null)) {
				header.setHeader("Transfer-Encoding", "chunked");
			}
			
			
			final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(header.getCharacterEncoding());
			
			IBodyWriter bodyWriter = new IBodyWriter() {
				
				public void flush(ByteBuffer[] bodyData) throws IOException {
					bodyDataSource.append(bodyData);
				}
				
				public void close() throws IOException {
					bodyDataSource.setComplete(true);
				}
				
				public void destroy() {
					bodyDataSource.destroy();
				}
			};
			BodyDataSink bodyDataSink = new BodyDataSink(null, header, bodyWriter);
			
			send(new HttpResponse(header, bodyDataSource));
			
			return bodyDataSink;
		}

		
		public BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException, IllegalStateException {
			
			if (responseHandler == null) {
				LOG.warning("response will not been send, because no response handler is assigned");
				return new BodyDataSink(null, header, null);
			}
			
			if ((header.getTransferEncoding() != null) && (header.getTransferEncoding().equalsIgnoreCase("chunked"))) {
				LOG.warning("message header contains Transfer-Encoding: chunked. removing it because message should be sent with predefined length (Content-Length)");
				header.removeHeader("Transfer-Encoding");
			}
			
			if (header.getContentLength() == -1) {
				header.setContentLength(contentLength);
			}
			
			
			final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(header.getCharacterEncoding());
			
			IBodyWriter bodyWriter = new IBodyWriter() {
				
				public void flush(ByteBuffer[] bodyData) throws IOException {
					bodyDataSource.append(bodyData);
				}
				
				public void close() throws IOException {
					bodyDataSource.setComplete(true);
				}
				
				public void destroy() {
					bodyDataSource.destroy();
				}
			};
			BodyDataSink bodyDataSink = new BodyDataSink(null, header, bodyWriter);
			
			send(new HttpResponse(header, bodyDataSource));
			
			return bodyDataSink;
		}

		
		public void send(final IHttpResponse response) throws IOException, IllegalStateException {
			if (isResponseCommitted) {
				throw new IllegalStateException("response is already committed"); 
			}			
			isResponseCommitted = true;

			
			if (responseHandler == null) {
				LOG.warning("response will not been send, because no response handler is assigned");
				return;
			}
			
			
			if (responseHandlerInfo.isResponseHandlerInvokeOnMessageReceived()) {
				
				IBodyCompleteListener copmleteListener = new IBodyCompleteListener() {
					
					@Execution(Execution.NONTHREADED)
					public void onComplete() throws IOException {
						performOnResponse(response);
					}
				};
				response.getNonBlockingBody().addCompleteListener(copmleteListener);
				
			} else {
				performOnResponse(response);
			}
		}
	
		
		
		
		private void performOnResponse(final IHttpResponse response) throws IOException {
			if (responseHandlerInfo.isResponseHandlerMultithreaded()) {
				
				Runnable task = new Runnable() {
					
					public void run() {
						try {
							responseHandler.onResponse(response);
						} catch (IOException ioe) {
							if (LOG.isLoggable(Level.FINE)) {
								LOG.fine("error occured by performing on response on " + responseHandler);
							}
						}
					}
				};
				
				con.getWorkerpool().execute(task);
			} else {
				responseHandler.onResponse(response);
			}
		}
		
		
		
		public void sendError(int errorCode) throws IllegalStateException {
			if (isResponseCommitted) {
				throw new IllegalStateException("response is already committed"); 
			}
			isResponseCommitted = true;

			
		}
		
		public void sendError(int errorCode, String msg) throws IllegalStateException {
			if (isResponseCommitted) {
				throw new IllegalStateException("response is already committed"); 
			}
			isResponseCommitted = true;
			
		}
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			
			if (requestHeader.getContentLength() != -1) {
				LOG.warning("message header contains content length. removing it because message should be sent as chunked");
				requestHeader.removeHeader("Content-Length");
			}
			
			if (requestHeader.getMethod().equalsIgnoreCase("GET") || requestHeader.getMethod().equalsIgnoreCase("HEAD")) {
				throw new IOException(requestHeader.getMethod() + " is a bodyless request");
			}

			if ((requestHeader.getTransferEncoding() == null)) {
				requestHeader.setHeader("Transfer-Encoding", "chunked");
			}
			
			if (isInternalForward()) {
				return con.send(requestHeader, responseHandler);
				
			} else {
				return httpClient.send(requestHeader, responseHandler);
			}
		}
		
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			
			if ((requestHeader.getTransferEncoding() != null) && (requestHeader.getTransferEncoding().equalsIgnoreCase("chunked"))) {
				LOG.warning("message header contains Transfer-Encoding: chunked. removing it because message should be sent with predefined length (Content-Length)");
				requestHeader.removeHeader("Transfer-Encoding");
			}
			
			if (requestHeader.getMethod().equalsIgnoreCase("GET") || requestHeader.getMethod().equalsIgnoreCase("HEAD")) {
				throw new IOException(requestHeader.getMethod() + " is a bodyless request");
			}

			if (requestHeader.getContentLength() == -1) {
				requestHeader.setContentLength(contentLength);
			}
			
			if (isInternalForward()) {
				return con.send(requestHeader, responseHandler);
			} else {
				return httpClient.send(requestHeader, responseHandler);
			}
		}
		
		
		public void forward(IHttpRequest request, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException {
			
			if (isInternalForward()) {
				con.send(request, responseHandler);
				
			} else {
				con.close();
				httpClient.send(request, responseHandler);
			}
		}	
		
		boolean isResponseCommitted() {
			return isResponseCommitted;
		}
		
		
		private boolean isInternalForward() {
			String url = request.getTargetURL().toString();
			if (url.indexOf("?") != -1) {
				url = url.substring(0, url.indexOf("?"));
			}

			if (url.equals(targetURL)) {
				return true;
			} else {
				return false;
			}
		}
	}
    
}
