/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.WritableByteChannel;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.IDataSink;
import org.xsocket.connection.AbstractNonBlockingStream;




/**
 * 
 * I/O resource capable of receiving the body data.
 * 
 * @author grro
 *
 */
public final class BodyDataSink extends AbstractNonBlockingStream implements IDataSink, WritableByteChannel, Closeable, Flushable {
	
	private static final Logger LOG = Logger.getLogger(BodyDataSink.class.getName());
	
	
	private final ArrayList<IBodyCloseListener> closeListeners = new ArrayList<IBodyCloseListener>();
	
	private IHttpHeader header = null;
	private AbstractHttpConnection httpConnection = null;
	
	private IBodyWriter bodySerializer = null;
	
	// close flag
	private boolean isOpen = true;
	
	
	BodyDataSink(AbstractHttpConnection httpConnection, IHttpHeader header, IBodyWriter bodySerializer) throws IOException {
		this.httpConnection = httpConnection;
		this.header = header;
		this.bodySerializer = bodySerializer;
		
		
		if (httpConnection != null) {
			httpConnection.setBodyDataSink(this);
			setFlushmode(httpConnection.getFlushmode());
		}

		setEncoding(header.getCharacterEncoding());

	}

	IBodyWriter getBodyWriter() {
		return bodySerializer;
	}
	
	void setBodyWriter(IBodyWriter bodyWriter) {
		this.bodySerializer = bodyWriter;
	}

	
	public void addCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			closeListeners.add(closeListener);
		}
	}

	
	
	public boolean removeCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			return closeListeners.remove(closeListener);
		}
	}
	
	
	final IHttpHeader getHeader() {
		return header;
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public void flush() {
		
		removeWriteMark();
		if (httpConnection != null) {
			httpConnection.setFlushmode(getFlushmode());
		}
		
		try {
			bodySerializer.flush(drainWriteQueue());
		} catch (IOException ioe) {
			throw new RuntimeException(ioe.toString());
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onWriteDataInserted() throws IOException, ClosedChannelException {
		
		if (isAutoflush()) {
			try {
				flush();
			} catch (Exception e) {
				throw new IOException(e.toString());
			}
		}
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final void close() throws IOException {

		if (isOpen) {
			isOpen = false;
			
			if (httpConnection != null) {
				httpConnection.removeBodyDataSink(this);
			}
			
			try {
				flush();
			} catch (Exception e) {
				throw new IOException(e.toString());
				
			} finally {
				bodySerializer.close();	
				callCloseListener();
			}
		}
	}
	
	
	@SuppressWarnings("unchecked")
	private void callCloseListener() {
		ArrayList<IBodyCloseListener> closeListenersCopy = null;
		synchronized (closeListeners) {
			closeListenersCopy = (ArrayList<IBodyCloseListener>) closeListeners.clone();
		}
		
		
		for (IBodyCloseListener bodyCloseListener : closeListenersCopy) {
			removeCloseListener(bodyCloseListener);
			callCompleteListener(bodyCloseListener);
		}
	}
	

	
	public boolean isOpen() {
		if (httpConnection != null) {
			return isOpen && httpConnection.isOpen();
		} else {
			return isOpen;
		}
	}
	
	
	@Override
	protected boolean isDataWriteable() {
		return isOpen;
	}

	@Override
	protected boolean isMoreInputDataExpected() {
		return isOpen;
	}
	
	
	
	@SuppressWarnings("unchecked")
	void onUnderlyingHttpConnectionClosed() {
		//onClosed();
		try {
			close();
		} catch (IOException ignore) { }
	}
	
	
	
	
	private void callCompleteListener(final IBodyCloseListener listener) {
		
		Runnable task = new Runnable() {
			
			public void run() {
				try {
					listener.onClose();
				} catch (IOException ioe) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("Error occured by calling close listener " + listener + " " + ioe.toString());
					}
				}
			}
		};
			
		
		if (HttpUtils.isBodyCloseListenerMutlithreaded(listener)) {
			processMultiThreaded(task);
		} else {
			processNonThreaded(task);
		}
	}
	

	private void processMultiThreaded(Runnable task) {
		
		if (httpConnection != null) {
			httpConnection.processMultiThreaded(task);
		} else {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("httpconnection not set. perform body listener call back method non threaded");
			}
			task.run();
		}		
	}
	
	private void processNonThreaded(Runnable task) {
		
		if (httpConnection != null) {
			httpConnection.processNonThreaded(task);

		} else {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("httpconnection not set. perform body listener call back method non threaded");
			}
			task.run();
		}		
	}
	
	
	public final void destroy() {
		isOpen = false;
		
		bodySerializer.destroy();
		callCloseListener();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return printWriteBuffer(getEncoding());
	}
}
