/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.MaxWriteSizeExceededException;




/**
 *  
 * @author grro@xsocket.org
 */
final class BoundBodyWriter implements IBodyWriter {

	private static final Logger LOG = Logger.getLogger(BoundBodyWriter.class.getName());
	
	private boolean isHeaderWritten = false;
		
	private int contentLength = 0;
	private int written = 0;

	private IHttpHeader header = null;
	private AbstractHttpConnection httpConnection = null;

	
	public BoundBodyWriter(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		this.header = header;
		this.httpConnection = httpConnection;

		if (header.getContentLength() < 0) {
			throw new IOException("Content Length has to be set");
		}
		
		contentLength = header.getContentLength();
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public void flush(ByteBuffer[] bodyData) throws IOException {
		
		try {
			int dataWritten = 0;
			
			if (!isHeaderWritten) {
				isHeaderWritten = true;
				
				if (LOG.isLoggable(Level.FINER)) {
					LOG.finer("[" + httpConnection.getId() + "] sending " + header.toString());
				}
				
				dataWritten = httpConnection.writeHeader(header);			
							
				dataWritten += writeBody(bodyData);
				
			} else {
				dataWritten = writeBody(bodyData);
			}
	
			if (dataWritten > 0) {
				httpConnection.flush();
			}
			
		} catch (IOException ioe) {
			if (httpConnection != null) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by flushing bound data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
			
				destroy();
				throw ioe;
			}	
		}
	}
	
	
	private int writeBody(ByteBuffer[] bodyData) throws IOException {

		int dataWritten = 0;
		
		if (bodyData != null) {
			dataWritten = (int) httpConnection.write(bodyData);
			written += dataWritten;
		}

		if (written > contentLength) {
			throw new MaxWriteSizeExceededException();
		}
		
		return dataWritten;
	}
	
	
	public void close() {

		try {
			
			if (written != contentLength) {
				throw new IOException("content size is " + written + " byte, but content-length field declares " + contentLength + " bytes"); 
			}
		
		
		} catch (IOException ioe) {
			if (httpConnection != null) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by closing bound data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
			
				httpConnection.destroy();
			}					
		} 
	}
	
	public void destroy() {
		httpConnection.destroy();
	}
}