/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.MaxReadSizeExceededException;
import org.xsocket.connection.INonBlockingConnection;



/**
 * 
 * @author grro@xsocket.org
 */
final class ChunkedBodyParser extends AbstractBodyParser {

	
	private static final Logger LOG = Logger.getLogger(ChunkedBodyParser.class.getName());
	
	private static final int STATE_READING_CHUNK = 0;
	private static final int STATE_READING_TRAILER = 1;
	private static final int STATE_CLOSED = 2;
	
	
	private int state = STATE_READING_CHUNK;

	private List<String> trailerList = null;
	
	private int totalSize = 0;
	private int chunkSize = 0;
	private int remainingDataToRead = 0;
	
	private IHttpHeader header = null;
	
		

	ChunkedBodyParser(AbstractHttpConnection httpConnection, IHttpHeader header, NonBlockingBodyDataSource dataSource) throws IOException {
		super(httpConnection, dataSource);
		this.header = header;
	}
	
	
	@Override
	boolean isReceiving() {
		return (state != STATE_CLOSED);
	}
	
	
	@Override
	void parseBody(INonBlockingConnection connection, NonBlockingBodyDataSource dataSource) throws IOException {
		
		if (state == STATE_READING_CHUNK) {
			readChunk(connection, dataSource);
		
		} else { // state == STATE_READING_TRAILER
			readTrailer(connection);
		}
	}
	
	
	private void readChunk(INonBlockingConnection connection,  NonBlockingBodyDataSource dataSource) throws IOException, BufferUnderflowException, MaxReadSizeExceededException {
		
		/////////////////////////
		// read chunk size	
		if (remainingDataToRead == 0) {   // new chunk?
			// read chunk size 
			String lengthField = connection.readStringByDelimiter("\r\n").trim();
						
			try {
				chunkSize = Integer.parseInt(lengthField, 16);
				
			} catch (NumberFormatException nfe) {

				// chunk extension?
				if (lengthField.indexOf(";") != -1) {
					lengthField	= lengthField.substring(0, lengthField.indexOf(";"));
					chunkSize = Integer.parseInt(lengthField, 16);
				} else {
					throw new IOException("[" + connection.getId() + "] http protocol error. length field expected");
				}
			}

			
			totalSize += chunkSize;

			
			// lastChunk?
			if (chunkSize == 0) {
				state = STATE_READING_TRAILER;
				readTrailer(connection);
				return;
				
			// .. no
			} else {
				remainingDataToRead = chunkSize + 2;  // chunk data + CRLF
				
				
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + connection.getId() + "] reading chunk size=" + remainingDataToRead);
				}
			}
		}
		
		
		/////////////////////////////
		// read chunk data
		int availableBytes = connection.available();
		ByteBuffer[] data = null;
		
		// complete chunk available?
		if (availableBytes >= remainingDataToRead) {
			// read chunk data
			if (availableBytes > 2) {
				data = connection.readByteBufferByLength(remainingDataToRead - 2);
			}
			
			// read chunk CRLF
			String crlf = connection.readStringByLength(2); 
			assert crlf.equals("\r\n") : "got " + crlf + " instead of chunk CRLF";
			
			remainingDataToRead = 0;
			

			if (data != null) {
				dataSource.append(data);
			}
			
			// read next chunk 
			readChunk(connection, dataSource);
			
		} else {
			
			// read chunk data 
			if (availableBytes > 2) {
				data = connection.readByteBufferByLength(availableBytes - 2);
				remainingDataToRead -= (availableBytes - 2);
			}
			

			if (data != null) {
				dataSource.append(data);
			}
		}
	}
	
	
	private void readTrailer(INonBlockingConnection connection) throws IOException, BufferUnderflowException, MaxReadSizeExceededException {
		String line = connection.readStringByDelimiter("\r\n");

		
		// empty line -> message has been completed
		if (line.length() == 0) {
			state = STATE_CLOSED;
			
			if (trailerList != null) {
				String[] trailers = trailerList.toArray(new String[trailerList.size()]);
				String[] headerLines = AbstractMessageHeader.unfoldingHeaderlines(trailers);
				for (String headerLine : headerLines) {
					header.addHeaderLine(headerLine);
				}
			}

			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + connection.getId() + "] complete chunk message received (body size=" + totalSize + ")");
			}

			setComplete();
			
		// trailer available
		} else {
			getTrailerList().add(line);
		}
	}
	
	
	private List<String> getTrailerList() {
		if (trailerList == null) {
			trailerList = new ArrayList<String>();
		}
		
		return trailerList;
	}
}
