/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;




/**
 * 
 * Chunked implementation of a body data sink
 * 
 * 
 * @author grro@xsocket.org
 */
final class ChunkedBodyWriter implements IBodyWriter {
	
	private static final Logger LOG = Logger.getLogger(ChunkedBodyWriter.class.getName());
	
	private boolean isHeaderWritten = false;

	private int writtenData = 0;
	
	private IHttpHeader header = null;
	private AbstractHttpConnection httpConnection = null;
	
	private boolean isOpen = true;
	
	/**
	 * constructor
	 * 
	 * @param httpConnection   the http connection
	 * @param header           the message header
	 * @param encoding         the encoding to use
	 */
	ChunkedBodyWriter(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		this.header = header;
		this.httpConnection = httpConnection;		
	}
	
		
	
	private void writeChunk(ByteBuffer[] bodyData) throws IOException {
		int written = 0;
		
		
		// write header (if not already written)
		if (!isHeaderWritten) {
			isHeaderWritten = true;

			
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + httpConnection.getId() + "] sending header ");
			}
			
			written += httpConnection.writeHeader(header);
		}

		
		
		// write body (if content size larger than 0)
		int length = 0;
		
		if (bodyData != null) {
			for (ByteBuffer buffer : bodyData) {
				length += buffer.remaining();
			}
		}
		
		if (length > 0) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + httpConnection.getId() + "] writing chunk (size=" + length + ")");
			}
			
			// write length field
			written += httpConnection.write(Integer.toHexString(length) +
                                                 "\r\n");
			
			// write data 
			written += httpConnection.write(bodyData);
			written += httpConnection.write("\r\n");
			
			writtenData += length;
		}
		
		if (written > 0) {
			httpConnection.flush();
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public void flush(ByteBuffer[] bodyData) throws IOException {
		
		
		try {
			
			writeChunk(bodyData);
			
		} catch (IOException ioe) {
			if (httpConnection != null) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by flushing chunked data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
			
				destroy();
				throw ioe;
			}	
		}
	}
	

	
	public void close() {

		if (isOpen) {
			isOpen = false;
		
			try {
				
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + httpConnection.getId()+ "] closing chunked body by writing termination chunk (body size=" + writtenData + ")");
				}

				// write termination chunk
				httpConnection.write("0" + 
						                          "\r\n" +
						  	                      "\r\n");
				httpConnection.flush();
			
			
			} catch (IOException ioe) {
				if (httpConnection != null) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("error occured by closing chunked data sink. Destroying connection. reason " + DataConverter.toString(ioe));
					}
				
					httpConnection.destroy();
				}						
			}
		}
	}
	
	
	public void destroy() {
		httpConnection.destroy();
	}
}
