/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;




import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.BufferUnderflowException;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.Execution;
import org.xsocket.IDataSink;
import org.xsocket.ILifeCycle;
import org.xsocket.Resource;
import org.xsocket.connection.IConnectionScoped;
import org.xsocket.connection.IServer;
import org.xsocket.connection.http.server.HttpProtocolAdapter;




/** 
 * A HTTP utility class
 *
 * @author grro@xsocket.org
 */
public final class HttpUtils {
	
	private static final Logger LOG = Logger.getLogger(HttpUtils.class.getName());
	
	@SuppressWarnings("unchecked")
	private static final Map<Class, Boolean> bodyDataExecutionModeCache = newMapCache(25);
	
	@SuppressWarnings("unchecked")
	private static final Map<Class, Boolean> bodyCompleteListenerExecutionModeCache = newMapCache(25);
	
	@SuppressWarnings("unchecked")
	private static final Map<Class, Boolean> bodyCloseListenerExecutionModeCache = newMapCache(25);

	@SuppressWarnings("unchecked")
	private static final Map<Class, HttpHandlerInfo> httpHandlerInfoCache = HttpUtils.newMapCache(25);
	public static final HttpHandlerInfo EMPTY_HTTP_HANDLER_INFO = new HttpHandlerInfo(null);

	
	private static String versionInfo = null;
	
	
	private static Map<String, String> mimeTypeMap = null;

	
	
	private HttpUtils() { }
		
	
	
	/**
	 * validate, based on a leading int length field. The length field will be removed
	 * 
	 * @param connection     the connection
	 * @return the length 
	 * @throws IOException if an exception occurs
	 * @throws BufferUnderflowException if not enough data is available
	 */
	public static int validateSufficientDatasizeByIntLengthField(NonBlockingBodyDataSource stream) throws IOException, BufferUnderflowException {
		return validateSufficientDatasizeByIntLengthField(stream, true) ;
	}
		
	
	/**
	 * validate, based on a leading int length field, that enough data (getNumberOfAvailableBytes() >= length) is available. If not,
	 * an BufferUnderflowException will been thrown. 
	 * 
	 * @param connection         the connection
	 * @param removeLengthField  true, if length field should be removed
	 * @return the length 
	 * @throws IOException if an exception occurs
	 * @throws BufferUnderflowException if not enough data is available
	 */
	public static int validateSufficientDatasizeByIntLengthField(NonBlockingBodyDataSource stream, boolean removeLengthField) throws IOException, BufferUnderflowException {

		stream.resetToReadMark();
		stream.markReadPosition();
		
		// check if enough data is available
		int length = stream.readInt();
		if (stream.available() < length) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("insufficient data. require " + length + " got "  + stream.available());
			}
			throw new BufferUnderflowException();
	
		} else { 
			// ...yes, remove mark
			if  (!removeLengthField) {
				stream.resetToReadMark();
			}
			stream.removeReadMark();
			return length;
		}
	}  
	
	
	
	/**
	 * get the mime type file to extension map 
	 *
	 * @return the mime type file to extension map
	 */
	public synchronized static Map<String, String> getMimeTypeMapping() {
		
		if (mimeTypeMap == null) {
			
			Map<String, String> map = new HashMap<String, String>();
			mimeTypeMap = Collections.unmodifiableMap(map); 
			
			try {
				InputStreamReader isr = new InputStreamReader(HttpUtils.class.getResourceAsStream("/org/xsocket/connection/http/mime.types"));
				if (isr != null) {
					LineNumberReader lnr = new LineNumberReader(isr);
					String line = null;
					do {
						line = lnr.readLine().trim();
						if (line != null) {
							if (!line.startsWith("#")) {
								StringTokenizer st = new StringTokenizer(line);
								if (st.hasMoreTokens()) {
									String mimeType = st.nextToken();
									while (st.hasMoreTokens()) {
										String extension = st.nextToken();
										map.put(extension, mimeType);
											
										if (LOG.isLoggable(Level.FINER)) {
											LOG.finer("mapping " + extension + " -> " + mimeType + " added");
										}
									}
								} else {
									if (LOG.isLoggable(Level.FINE)) {
										LOG.fine("line " + line + "ignored");
									}	
								}
							}
						}
					} while (line != null);
		
					lnr.close();
				}
			} catch (Exception ignore) { }
		}
		
		return mimeTypeMap;
	}

	
	

	/**
	 * injects a server field
	 *
	 * @param handler   the handler
	 * @param server    the server to inject
	 */
	static void injectServerField(Object handler, IServer server) {
		Field[] fields = handler.getClass().getDeclaredFields();
		for (Field field : fields) {
			if (field.isAnnotationPresent(Resource.class)) {
				Resource res = field.getAnnotation(Resource.class);
				if ((field.getType() == IServer.class) || (res.type() == IServer.class)) {
					field.setAccessible(true);
					try {
						field.set(handler, server);
					} catch (IllegalAccessException iae) {
						LOG.warning("could not set HandlerContext for attribute " + field.getName() + ". Reason " + iae.toString());
					}
				}
			}
		}
	}


	/**
	 * inject a protocol adapter
	 *
	 * @param handler   the handler
	 * @param adapter   the adapter to inject
	 */
	static void injectProtocolAdapter(Object handler, HttpProtocolAdapter adapter) {
		Field[] fields = handler.getClass().getDeclaredFields();
		for (Field field : fields) {
			if (field.isAnnotationPresent(Resource.class)) {
				Resource res = field.getAnnotation(Resource.class);
				if ((field.getType() == HttpProtocolAdapter.class) || (res.type() == HttpProtocolAdapter.class)) {
					field.setAccessible(true);
					try {
						field.set(handler, adapter);
					} catch (IllegalAccessException iae) {
						LOG.warning("could not set HandlerContext for attribute " + field.getName() + ". Reason " + iae.toString());
					}
				}
			}
		}
	}


	static boolean isContentTypeFormUrlencoded(IHttpMessage message) {
		
		if (!message.hasBody()) {
			return false;
		}
		
		String contentType = message.getContentType();
		if ((contentType != null) && (contentType.startsWith("application/x-www-form-urlencoded"))) {
			return true;
		}
		
		return false;
	}
	
	
	static String getContentTypedFormUrlencodedEncodingType(IHttpMessage message) {
		
		String contentType = message.getContentType();
		if ((contentType != null) && (contentType.startsWith("application/x-www-form-urlencoded"))) {
			String[] parts = contentType.split(";");
			if (parts.length > 1) {
				for(int i = 1; i < parts.length; i++) {
					String[] kvp = parts[i].split("=");
					if (kvp[0].trim().toUpperCase().equals("CHARSET")) {
						return kvp[1].trim();
					}
				}
			}
		}
		
		return "UTF-8";
	}
	

	
	static IHttpRequest newFormEncodedRequestWrapper(IHttpRequest request) throws IOException {
		if (request instanceof HttpRequestWrapper) {
			return request;
		}
		
		return new HttpRequestWrapper(request);
	}
	
	

	@SuppressWarnings("unchecked")
	static HttpHandlerInfo getHttpHandlerInfo(IHttpHandler httpHandler) {
		if (httpHandler == null) {
			return EMPTY_HTTP_HANDLER_INFO;
		}

		HttpHandlerInfo httpHandlerInfo = httpHandlerInfoCache.get(httpHandler.getClass());

		if (httpHandlerInfo == null) {
			httpHandlerInfo = new HttpHandlerInfo((Class<IHttpHandler>) httpHandler.getClass());
			httpHandlerInfoCache.put(httpHandler.getClass(), httpHandlerInfo);
		}

		return httpHandlerInfo;
	}
	 

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseHandlerMultithreaded();
	}

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseExceptionHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseExeptionHandlerMultithreaded();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseHandlerInvokedOnMessageReceived(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseHandlerInvokeOnMessageReceived();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isRequestHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isRequestHandlerMultithreaded();
	}

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isRequestHandlerInvokedOnMessageReceived(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isRequestHandlerInvokeOnMessageReceived();
	}

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isRequestHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isRequestHandler();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseTimeoutHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseTimeoutHandler();
	}

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseTimeoutHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseTimeoutHandlerMultithreaded();
	}


	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseHandlerInvokeOnMessageReceived(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseHandlerInvokeOnMessageReceived();
	}
	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isResponseHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isResponseHandler();
	}


	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isConnectHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isConnectHandler();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isConnectHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isConnectHandlerMultithreaded();
	}

	

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isDisconnectHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isDisconnectHandler();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isDisconnectHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isDisconnectHandlerMultithreaded();
	}

	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isRequestTimeoutHandler(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isRequestTimeoutHandler();
	}

	
	/**
	 * <b>This is a xSocket internal method and subject to change</b>
	 */
	public static boolean isRequestTimeoutHandlerMultithreaded(IHttpHandler handler) {
		return getHttpHandlerInfo(handler).isRequestTimeoutHandlerMultithreaded();
	}


	
	
	/**
	 * returns the execution mode for the given body handler 
	 * 
	 * @param bodyHandler  the body handler 
	 * @return the execution mode
	 */
	static boolean isMutlithreaded(IBodyDataHandler bodyHandler) {
		Boolean isMutlithreaded = bodyDataExecutionModeCache.get(bodyHandler.getClass());
		
		if (isMutlithreaded == null) {
			int mode = IBodyDataHandler.DEFAULT_EXECUTION_MODE;
			
			Execution execution = bodyHandler.getClass().getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
			
			try {
				Method meth = bodyHandler.getClass().getMethod("onData", new Class[] { NonBlockingBodyDataSource.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}
				
			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}
			
			isMutlithreaded = (mode == Execution.MULTITHREADED);
			
			bodyDataExecutionModeCache.put(bodyHandler.getClass(), isMutlithreaded);
		}

		return isMutlithreaded;
	}

	
	
	/**
	 * returns the execution mode for the given body complete listener 
	 * 
	 * @param bodyHandler  the body handler 
	 * @return the execution mode
	 */
	static boolean isBodyCompleteListenerMutlithreaded(IBodyCompleteListener completeListener) {
		Boolean isMutlithreaded = bodyCompleteListenerExecutionModeCache.get(completeListener.getClass());
		 
		if (isMutlithreaded == null) {
			int mode = IBodyDataHandler.DEFAULT_EXECUTION_MODE;
			
			Execution execution = completeListener.getClass().getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
			
			try {
				Method meth = completeListener.getClass().getMethod("onComplete", new Class[] { });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}
				
			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}
			
			isMutlithreaded = (mode == Execution.MULTITHREADED);
			
			bodyCompleteListenerExecutionModeCache.put(completeListener.getClass(), isMutlithreaded);
		}

		return isMutlithreaded;
	}
	

	
	/**
	 * returns the execution mode for the given body complete listener 
	 * 
	 * @param bodyHandler  the body handler 
	 * @return the execution mode
	 */
	static boolean isBodyCloseListenerMutlithreaded(IBodyCloseListener closeListener) {
		Boolean isMutlithreaded = bodyCloseListenerExecutionModeCache.get(closeListener.getClass());
		 
		if (isMutlithreaded == null) {
			int mode = IBodyDataHandler.DEFAULT_EXECUTION_MODE;
			
			Execution execution = closeListener.getClass().getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
			
			try {
				Method meth = closeListener.getClass().getMethod("onClose", new Class[] { });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}
				
			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}
			
			isMutlithreaded = (mode == Execution.MULTITHREADED);
			
			bodyCloseListenerExecutionModeCache.put(closeListener.getClass(), isMutlithreaded);
		}

		return isMutlithreaded;
	}


	
	/**
	 * get the version info
	 * 
	 * @return the version info
	 */
	public static String getVersionInfo() {
		
		if (versionInfo == null) {
			
			versionInfo = "<unknown>";
			
			try {
				InputStreamReader isr = new InputStreamReader(HttpUtils.class.getResourceAsStream("/org/xsocket/connection/http/version.txt"));
				if (isr != null) {
					LineNumberReader lnr = new LineNumberReader(isr);
					String line = null;
					do {
						line = lnr.readLine();
						if (line != null) {
							if (line.startsWith("Implementation-Version=")) {
								versionInfo = line.substring("Implementation-Version=".length(), line.length()).trim();
							}
						}
					} while (line != null);
		
					lnr.close();
				}
			} catch (Exception ignore) { }
		} 
		
		return versionInfo;
	}
	
	

	
	/**
	 * create a new synchronized map cache
	 *  
	 * @param <T>       the map key type
	 * @param <E>       the map value type
	 * @param maxSize   the max cache size (oldest elements will be removed by exceeding this size)
	 * @return the ne wcache
	 */
	public static <T, E>  Map<T, E> newMapCache(int maxSize) {
		return Collections.synchronizedMap(new Cache<T, E>(maxSize));
	}
	

	@SuppressWarnings("unchecked")
	private static final class Cache<T, E> extends LinkedHashMap<T, E> {
		
		private static final long serialVersionUID = 4513864504007457500L;
		
		private int maxSize = 0;
		
		Cache(int maxSize) {
			this.maxSize = maxSize;
		}
		

		@Override
		protected boolean removeEldestEntry(Entry<T, E> eldest) {
			return size() > maxSize;
		}	
	}
	
	
	static final class HttpHandlerInfo {
		
		private boolean isRequestHandler = false;
		private boolean isRequestHandlerInvokeOnMessageReceived = false;;
		private boolean isRequestHandlerMultithreaded = true;

		
		private boolean isResponseHandler = false;
		private boolean isResponseHandlerInvokeOnMessageReceived = false;;
		private boolean isResponseHandlerMultithreaded = true;
		private boolean isResponseExeptionHandlerMultithreaded = true;
		
		private boolean isResponseTimeoutHandler = false;
		private boolean isResponseTimeoutHandlerMultithreaded = true;
		
		private boolean isConnectHandler = false;
		private boolean isConnectHandlerMultithreaded = true;

		
		private boolean isDisconnectHandler = false;
		private boolean isDisconnectHandlerMultithreaded = true;

		private boolean isRequestTimoutHandler = false;
		private boolean isRequestTimoutHandlerMultithreaded = true;

		
		private boolean isConnectionScoped = false;
		private boolean isLifeCycle = false;
		
		

		@SuppressWarnings("unchecked")
		public HttpHandlerInfo(Class clazz) {

			if (clazz == null) {
				return;
			}
			
			isConnectionScoped = IConnectionScoped.class.isAssignableFrom(clazz);
			isLifeCycle = ILifeCycle.class.isAssignableFrom(clazz);

			if (IHttpRequestHandler.class.isAssignableFrom(clazz)) {
				isRequestHandler = true;
				isRequestHandlerMultithreaded = isOnRequestMultithreaded((Class<IHttpRequestHandler>) clazz);
				isRequestHandlerInvokeOnMessageReceived = isOnRequestInvokeOnMessageReceived((Class<IHttpRequestHandler>) clazz);
			}
			
			if (IHttpResponseHandler.class.isAssignableFrom(clazz)) {
				isResponseHandler = true;
				isResponseHandlerMultithreaded = isOnResponseMultithreaded((Class<IHttpResponseHandler>) clazz);
				isResponseHandlerInvokeOnMessageReceived = isOnResponseInvokeOnMessageReceived((Class<IHttpResponseHandler>) clazz);
				
				isResponseExeptionHandlerMultithreaded = isOnResponseExceptionMultithreaded((Class<IHttpResponseHandler>) clazz);
			}

			if (IHttpResponseTimeoutHandler.class.isAssignableFrom(clazz)) {
				isResponseTimeoutHandler = true;
				isResponseTimeoutHandlerMultithreaded = isOnResponseTimeoutMultithreaded((Class<IHttpResponseHandler>) clazz);

			}

			
			if (IHttpConnectHandler.class.isAssignableFrom(clazz)) {
				isConnectHandler = true;
				isConnectHandlerMultithreaded = isOnConnectMultithreaded(clazz);
			}

			if (IHttpDisconnectHandler.class.isAssignableFrom(clazz)) {
				isDisconnectHandler = true;
				isDisconnectHandlerMultithreaded = isOnDisconnectMultithreaded(clazz);
			}

			if (IHttpRequestTimeoutHandler.class.isAssignableFrom(clazz)) {
				isRequestTimoutHandler = true;
				isRequestTimoutHandlerMultithreaded = isOnRequestTimeoutMultithreaded(clazz);
			}
		}


		static boolean isOnRequestMultithreaded(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = serverHandlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
 
			try {
				Method meth = serverHandlerClass.getMethod("onRequest", new Class[] { IHttpExchange.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}

		
		static boolean isOnResponseMultithreaded(Class<IHttpResponseHandler> handlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = handlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
 
			try {
				Method meth = handlerClass.getMethod("onResponse", new Class[] { IHttpResponse.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}

		
		static boolean isOnResponseExceptionMultithreaded(Class<IHttpResponseHandler> handlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = handlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
 
			try {
				Method meth = handlerClass.getMethod("onException", new Class[] { IOException.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}
		

		static boolean isOnResponseTimeoutMultithreaded(Class<IHttpResponseHandler> handlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = handlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}
 
			try {
				Method meth =handlerClass.getMethod("onException", new Class[] { SocketTimeoutException.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}
		
		
		static boolean isOnRequestInvokeOnMessageReceived(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_INVOKE_ON_MODE;

			InvokeOn invokeOn = serverHandlerClass.getAnnotation(InvokeOn.class);
			if (invokeOn != null) {
				mode = invokeOn.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onRequest", new Class[] { IHttpExchange.class });
				invokeOn = meth.getAnnotation(InvokeOn.class);
				if (invokeOn != null) {
					mode = invokeOn.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because response handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == InvokeOn.MESSAGE_RECEIVED);
		}


		static boolean isOnResponseInvokeOnMessageReceived(Class<IHttpResponseHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_INVOKE_ON_MODE;

			InvokeOn invokeOn = serverHandlerClass.getAnnotation(InvokeOn.class);
			if (invokeOn != null) {
				mode = invokeOn.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onResponse", new Class[] { IHttpResponse.class });
				invokeOn = meth.getAnnotation(InvokeOn.class);
				if (invokeOn != null) {
					mode = invokeOn.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because response handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == InvokeOn.MESSAGE_RECEIVED);
		}




		static boolean isOnConnectMultithreaded(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = serverHandlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onConnect", new Class[] { IHttpConnection.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}


		static boolean isOnDisconnectMultithreaded(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = serverHandlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onDisconnect", new Class[] { IHttpConnection.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}


		static boolean isOnRequestTimeoutMultithreaded(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = serverHandlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onRequestTimeout", new Class[] { IHttpConnection.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}


		static boolean isOnConnectionTimeoutMultithreaded(Class<IHttpRequestHandler> serverHandlerClass) {
			int mode = IHttpRequestHandler.DEFAULT_EXECUTION_MODE;

			Execution execution = serverHandlerClass.getAnnotation(Execution.class);
			if (execution != null) {
				mode = execution.value();
			}

			try {
				Method meth = serverHandlerClass.getMethod("onConnectionTimeout", new Class[] { IHttpConnection.class });
				execution = meth.getAnnotation(Execution.class);
				if (execution != null) {
					mode = execution.value();
				}

			} catch (NoSuchMethodException nsme) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("shouldn't occure because body handler has to have such a method " + nsme.toString());
				}
			}

			return (mode == Execution.MULTITHREADED);
		}

		public boolean isConnectHandler() {
			return isConnectHandler;
		}

		public boolean isConnectHandlerMultithreaded() {
			return isConnectHandlerMultithreaded;
		}

		public boolean isDisconnectHandler() {
			return isDisconnectHandler;
		}

		public boolean isDisconnectHandlerMultithreaded() {
			return isDisconnectHandlerMultithreaded;
		}
	
		public boolean isRequestTimeoutHandler() {
			return isRequestTimoutHandler;
		}
		
		public boolean isRequestTimeoutHandlerMultithreaded() {
			return isRequestTimoutHandlerMultithreaded;
		}
		
		public boolean isConnectionScoped() {
			return isConnectionScoped;
		}
		
		public boolean isLifeCycle() {
			return isLifeCycle;
		}
		
		public boolean isRequestHandler() {
			return isRequestHandler;
		}
		
		public boolean isRequestHandlerInvokeOnMessageReceived() {
			return  isRequestHandlerInvokeOnMessageReceived;
		}

		public boolean isRequestHandlerMultithreaded() {
			return isRequestHandlerMultithreaded;
		}
		
		public boolean isResponseHandler() {
			return isResponseHandler;
		}
		
		public boolean isResponseHandlerInvokeOnMessageReceived() {
			return  isResponseHandlerInvokeOnMessageReceived;
		}

		public boolean isResponseHandlerMultithreaded() {
			return isResponseHandlerMultithreaded;
		}
		
		public boolean isResponseTimeoutHandler() {
			return isResponseTimeoutHandler;
		}

		public boolean isResponseExeptionHandlerMultithreaded() {
			return isResponseExeptionHandlerMultithreaded;
		}
		

		public boolean isResponseTimeoutHandlerMultithreaded() {
			return isResponseTimeoutHandlerMultithreaded;
		}
	}
	
	

	
	
	private static final class HttpRequestWrapper implements IHttpRequest {
		
		private IHttpRequest delegee = null;
		private HttpRequestHeaderWrapper headerWrapper = null;
		
		public HttpRequestWrapper(IHttpRequest delegee) throws IOException {
			this.delegee = delegee;
			headerWrapper = new HttpRequestHeaderWrapper(delegee);
		}
		
		public void setContentLength(int length) {
			headerWrapper.setContentLength(length);
		}
		
		public int getContentLength() {
			return headerWrapper.getContentLength();
		}
		
		public void setContentType(String type) {
			headerWrapper.setContentType(type);
		}
		
		public String getContentType() {
			return headerWrapper.getContentType();
		}
		
		public String getCharacterEncoding() {
			return headerWrapper.getCharacterEncoding();
		}
		
		public void setTransferEncoding(String transferEncoding) {
			headerWrapper.setTransferEncoding(transferEncoding);
		}
		
		public String getTransferEncoding() {
			return headerWrapper.getTransferEncoding();
		}
		
		public IHttpRequestHeader getRequestHeader() {
			return headerWrapper;
		}
		
		public String getHeader(String headername) {
			return headerWrapper.getHeader(headername);
		}

		public void setHeader(String headername, String headervalue) {
			headerWrapper.setHeader(headername, headervalue);
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getHeaderNames() {
			return headerWrapper.getHeaderNames();
		}
		
		public List<String> getHeaderList(String headername) {
			return headerWrapper.getHeaderList(headername);
		}
		
		public Set<String> getHeaderNameSet() {
			return headerWrapper.getHeaderNameSet();
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getHeaders(String headername) {
			return headerWrapper.getHeaders(headername);
		}
		
		public boolean containsHeader(String headername) {
			return headerWrapper.containsHeader(headername);
		}
		
		public boolean containsHeaderValue(String headername, String headervalue) {
			return headerWrapper.containsHeaderValue(headername, headervalue);
		}
		
		public void addHeader(String headername, String headervalue) {
			headerWrapper.addHeader(headername, headervalue);
		}
		
		public void addHeaderLine(String line) {
			headerWrapper.addHeaderLine(line);
		}
		
		public void removeHeader(String headername) {
			headerWrapper.removeHeader(headername);
		}
		
		public void removeHopByHopHeaders() {
			headerWrapper.removeHopByHopHeaders();
		}
		
		public void removeHopByHopHeaders(String... headernames) {
			headerWrapper.removeHopByHopHeaders(headernames);
		}
		
		public Boolean getBooleanParameter(String name) {
			return headerWrapper.getBooleanParameter(name);
		}
		
		public boolean getBooleanParameter(String name, boolean defaultVal) {
			return headerWrapper.getBooleanParameter(name, defaultVal);
		}
		
		public Double getDoubleParameter(String name) {
			return headerWrapper.getDoubleParameter(name);
		}
		
		public double getDoubleParameter(String name, double defaultVal) {
			return headerWrapper.getDoubleParameter(name, defaultVal);
		}
		
		public Float getFloatParameter(String name) {
			return headerWrapper.getFloatParameter(name);
		}
		
		public float getFloatParameter(String name, float defaultVal) {
			return headerWrapper.getFloatParameter(name, defaultVal);
		}
		
		public Integer getIntParameter(String name) {
			return headerWrapper.getIntParameter(name);
		}
		
		public int getIntParameter(String name, int defaultVal) {
			return headerWrapper.getIntParameter(name, defaultVal);
		}
		
		public Long getLongParameter(String name) {
			return headerWrapper.getLongParameter(name);
		}
		
		public long getLongParameter(String name, long defaultVal) {
			return headerWrapper.getLongParameter(name, defaultVal);
		}
		
		public String getMethod() {
			return headerWrapper.getMethod();
		}
		
		public void setMethod(String method) {
			headerWrapper.setMethod(method);			
		}
		
		public String getParameter(String name) {
			return headerWrapper.getParameter(name);
		}
		
		public void setParameter(String parameterName, String parameterValue) {
			headerWrapper.setParameter(parameterName, parameterValue);
		}

		@SuppressWarnings("unchecked")
		public Map getParameterMap() {
			return headerWrapper.getParameterMap();
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getParameterNames() {
			return headerWrapper.getParameterNames();
		}
		
		public Set<String> getParameterNameSet() {
			return headerWrapper.getParameterNameSet();
		}
		
		public String getProtocol() {
			return headerWrapper.getProtocol();
		}
		
		public String getQueryString() {
			return headerWrapper.getQueryString();
		}
		
		public String getRemoteHost() {
			return headerWrapper.getRemoteHost();
		}
		
		public int getRemotePort() {
			return headerWrapper.getRemotePort();
		}
		
		public String getRemoteAddr() {
			return headerWrapper.getRemoteAddr();
		}
		
		public String getRequestURI() {
			return headerWrapper.getRequestURI();
		}
		
		public String getScheme() {
			return headerWrapper.getScheme();
		}
		
		public void setScheme(String scheme) {
			headerWrapper.setScheme(scheme);
		}
		
		public String getServerName() {
			return headerWrapper.getServerName();
		}
		
		public int getServerPort() {
			return headerWrapper.getServerPort();
		}

		public URL getTargetURL() {
			return headerWrapper.getTargetURL();
		}
		
		public void setTargetURL(URL url) {
			headerWrapper.setTargetURL(url);
		}
		
		public void updateTargetURL(String host, int port) {
			headerWrapper.updateTargetURL(host, port);
		}
		
		public boolean isSecure() {
			return headerWrapper.isSecure();
		}
		
		
		public String getContextPath() {
			return delegee.getContextPath();
		}
		
		public void setContextPath(String contextPath) {
			delegee.setContextPath(contextPath);
		}
		
		public String getRequestHandlerPath() {
			return delegee.getRequestHandlerPath();
		}
		
		public void setRequestHandlerPath(String requestHandlerPath) {
			delegee.setRequestHandlerPath(requestHandlerPath);
		}
		
		public boolean hasBody() {
			return delegee.hasBody();
		}
		
		public BlockingBodyDataSource getBlockingBody() throws IOException {
			return delegee.getBlockingBody();
		}
		
		public NonBlockingBodyDataSource getNonBlockingBody() throws IOException {
			return delegee.getNonBlockingBody();
		}
		
		@Override
		public String toString() {
			return delegee.toString();
		}
	}
	
	
	private static final class HttpRequestHeaderWrapper implements IHttpRequestHeader {
		
		private IHttpRequestHeader delegee = null;
		private Map<String, String> bodyParamsMap = new HashMap<String, String>();
		
		HttpRequestHeaderWrapper(IHttpRequest request) throws IOException {
			this.delegee = request.getRequestHeader();
			
			if (HttpUtils.isContentTypeFormUrlencoded(request) && request.hasBody()) {
				bodyParamsMap.putAll(parseFormParamters(request.getNonBlockingBody().toString(), HttpUtils.getContentTypedFormUrlencodedEncodingType(request)));
			}
		}
		
		public void setContentLength(int length) {
			delegee.setContentLength(length);
		}
		
		public int getContentLength() {
			return delegee.getContentLength();
		}
		
		public void setContentType(String type) {
			delegee.setContentType(type);
		}
		
		public String getContentType() {
			return delegee.getContentType();
		}
		
		public String getCharacterEncoding() {
			return delegee.getCharacterEncoding();
		}
		
		public void setTransferEncoding(String transferEncoding) {
			delegee.setTransferEncoding(transferEncoding);
		}
		
		public String getTransferEncoding() {
			return delegee.getTransferEncoding();
		}
		
		
		public String getHeader(String headername) {
			return delegee.getHeader(headername);
		}

		public void setHeader(String headername, String headervalue) {
			delegee.setHeader(headername, headervalue);
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getHeaderNames() {
			return delegee.getHeaderNames();
		}
		
		public List<String> getHeaderList(String headername) {
			return delegee.getHeaderList(headername);
		}
		
		public Set<String> getHeaderNameSet() {
			return delegee.getHeaderNameSet();
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getHeaders(String headername) {
			return delegee.getHeaders(headername);
		}
		
		public boolean containsHeader(String headername) {
			return delegee.containsHeader(headername);
		}
		
		public boolean containsHeaderValue(String headername, String headervalue) {
			return delegee.containsHeaderValue(headername, headervalue);
		}
		
		public void addHeader(String headername, String headervalue) {
			delegee.addHeader(headername, headervalue);
		}
		
		public void addHeaderLine(String line) {
			delegee.addHeaderLine(line);
		}
		
		public void removeHeader(String headername) {
			delegee.removeHeader(headername);
		}
		
		public void removeHopByHopHeaders() {
			delegee.removeHopByHopHeaders();
		}
		
		public void removeHopByHopHeaders(String... headernames) {
			delegee.removeHopByHopHeaders(headernames);
		}
		

		public Integer getIntParameter(String name) {
			String s = getParameter(name);
			if (s != null) {
				return Integer.parseInt(s);
			} else {
				return null;
			}
		}
		


		public int getIntParameter(String name, int defaultVal) {
			String s = getParameter(name);
			if (s != null) {
				try {
					return Integer.parseInt(s);
				} catch (Exception e) {
					return defaultVal;
				}
			} else {
				return defaultVal;
			}
		}
		

		public Long getLongParameter(String name) {
			String s = getParameter(name);
			if (s != null) {
				return Long.parseLong(s);
			} else {
				return null;
			}
		}
		

		public long getLongParameter(String name, long defaultVal) {
			String s = getParameter(name);
			if (s != null) {
				try {
					return Long.parseLong(s);
				} catch (Exception e) {
					return defaultVal;
				}
			} else {
				return defaultVal;
			}
		}


		public Double getDoubleParameter(String name) {
			String s = getParameter(name);
			if (s != null) {
				return Double.parseDouble(s);
			} else {
				return null;
			}
		}
		

		public double getDoubleParameter(String name, double defaultVal) {
			String s = getParameter(name);
			if (s != null) {
				try {
					return Double.parseDouble(s);
				} catch (Exception e) {
					return defaultVal;
				}
			} else {
				return defaultVal;
			}
		}
		
		
		public Float getFloatParameter(String name) {
			String s = getParameter(name);
			if (s != null) {
				return Float.parseFloat(s);
			} else {
				return null;
			}
		}
		
		public float getFloatParameter(String name, float defaultVal) {
			String s = getParameter(name);
			if (s != null) {
				try {
					return Float.parseFloat(s);
				} catch (Exception e) {
					return defaultVal;
				}
			} else {
				return defaultVal;
			}
		}
		
		public Boolean getBooleanParameter(String name) {
			String s = getParameter(name);
			if (s != null) {
				return Boolean.parseBoolean(s);
			} else {
				return null;
			}
		}
		
		public boolean getBooleanParameter(String name, boolean defaultVal) {
			String s = getParameter(name);
			if (s != null) {
				try {
					return Boolean.parseBoolean(s);
				} catch (Exception e) {
					return defaultVal;
				}
			} else {
				return defaultVal;
			}
		}
				
		public String getMethod() {
			return delegee.getMethod();
		}
		
		public void setMethod(String method) {
			delegee.setMethod(method);			
		}
		
		public String getParameter(String name) {
			if (bodyParamsMap.containsKey(name)) {
				return bodyParamsMap.get(name);
			}
			
			return delegee.getParameter(name);
		}
		
		
		public void setParameter(String parameterName, String parameterValue) {
			if (bodyParamsMap.containsKey(parameterName)) {
				throw new RuntimeException("parameter is contained in body and can not be modified");
			}
			
			delegee.setParameter(parameterName, parameterValue);
		}

		@SuppressWarnings("unchecked")
		public Map getParameterMap() {
			Map result = new HashMap();
			
			result.putAll(delegee.getParameterMap());
			result.putAll(bodyParamsMap);
			
			return result;
		}
		
		
		public Set<String> getParameterNameSet() {
			Set<String> result = new HashSet<String>();
			
			result.addAll(delegee.getParameterNameSet());
			result.addAll(bodyParamsMap.keySet());
			
			return result;
		}
		
		@SuppressWarnings("unchecked")
		public Enumeration getParameterNames() {
			return Collections.enumeration(getParameterNameSet());
		}
		
		
		public String getProtocol() {
			return delegee.getProtocol();
		}
		
		public String getQueryString() {
			return delegee.getQueryString();
		}
		
		public String getRemoteHost() {
			return delegee.getRemoteHost();
		}
		
		public int getRemotePort() {
			return delegee.getRemotePort();
		}
		
		public String getRemoteAddr() {
			return delegee.getRemoteAddr();
		}
		
		public String getRequestURI() {
			return delegee.getRequestURI();
		}
		
		public String getScheme() {
			return delegee.getScheme();
		}
		
		public void setScheme(String scheme) {
			delegee.setScheme(scheme);
		}
		
		public String getServerName() {
			return delegee.getServerName();
		}
		
		public int getServerPort() {
			return delegee.getServerPort();
		}

		public URL getTargetURL() {
			return delegee.getTargetURL();
		}
		
		public void setTargetURL(URL url) {
			delegee.setTargetURL(url);
		}
		
		public void updateTargetURL(String host, int port) {
			delegee.updateTargetURL(host, port);
		}
		
		public boolean isSecure() {
			return delegee.isSecure();
		}
		
	
		public int writeTo(IDataSink dataSink) throws IOException {
			return delegee.writeTo(dataSink);
		}
	
		@Override
		public String toString() {
			return delegee.toString();
		}
		
		private static Map<String, String> parseFormParamters(String body, String encoding) throws IOException {
			Map<String, String> result = new HashMap<String, String>();
			
			String[] params = body.split("&");
			for (String param : params) {
				String[] kv = param.split("=");
				if (kv.length > 1) {
					result.put(kv[0], URLDecoder.decode(kv[1], encoding));
				}
			}
			
			return result;
		}
	}
}
