/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.net.ConnectException;





/**
 * This class encapsulates a HTTP request received and a
 * response to be generated in one exchange. It provides methods 
 * <ul>
 *   <li>to retrieve and forward the the request, and</li>
 *   <li>to send the response</li>
 * </ul>
 *
 *
 * @author grro@xsocket.org
 */
public interface IHttpExchange {
	
	
	/**
	 * get the request of this exchange
	 * 
	 * @return the request
	 */
	public IHttpRequest getRequest();

	
	
	/**
	 * returns the underlying connection
	 *
	 * @return the connection
	 */
	public IHttpConnection getConnection();
	
	
	

	/**
	 * send the response. If the content-length header is not set, the message will send 
	 * in chunked mode 
	 *
	 * @param header   the header
	 * @return the body handle to write
	 * @throws IOException if an exception occurs
	 * @throws IllegalStateException if a response has already been send 
	 */
	public BodyDataSink send(IHttpResponseHeader header) throws IOException, IllegalStateException;
	
	
	/**
	 * send the response in a plain body non-blocking mode
	 *
	 * @param header         the header
	 * @param contentLength  the body content length
	 * @return the body handle to write
	 * @throws IOException if an exception occurs
	 * @throws IllegalStateException if a response has already been send 
	 */
	public BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException, IllegalStateException;
	

	/**
	 * send the response
	 *
	 * @param response   the response
	 * @throws IOException if an exception occurs
	 * @throws IllegalStateException if a response has already been send 
	 */
	public void send(IHttpResponse response) throws IOException, IllegalStateException;
	

	/**
	 * send an error response
	 *
	 * @param errorCode   the error code
	 * @param msg         the error message
	 * @throws IllegalStateException if a response has already been send 
	 */
	public void sendError(int errorCode, String msg) throws IllegalStateException;

	
	/**
	 * send an error response
	 *
	 * @param errorCode   the error code
	 * @param msg         the error message
	 * @throws IllegalStateException if a response has already been send 
	 */
	public void sendError(int errorCode) throws IllegalStateException;


	/**
	 * forwards the a request. If the content-length header is not set, the message will send 
	 * in chunked mode 
	 *
	 * @param requestHeader    the request header
	 * @param responseHandler  the response handler (supported: {@link IHttpRequestHandler}, {@link IHttpResponseTimeoutHandler})
	 * @return the body handle to write
	 * @throws ConnectException if an error occurred while attempting to connect to a remote address and port. 
	 * @throws IOException if an exception occurs	
	 * @throws IllegalStateException if a request has already been forwarded
	 */
	public BodyDataSink forward(IHttpRequestHeader requestHeader, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException;
 

	/**
	 * forwards a request. The request will be send in a plain body mode
	 *
	 * @param requestHeader     the request header
	 * @param contentLength     the content length
	 * @param responseHandler   the response handler (supported: {@link IHttpRequestHandler}, {@link IHttpResponseTimeoutHandler})
	 * @return the body handle to write
	 * @throws ConnectException if an error occurred while attempting to connect to a remote address and port. 
	 * @throws IOException if an exception occurs
	 * @throws IllegalStateException if a request has already been forwarded 
	 */	 
	public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException;


	/**
	 * forwards a request 
	 *  
	 * @param request           the request
	 * @param responseHandler   the response handler  (supported: {@link IHttpRequestHandler}, {@link IHttpResponseTimeoutHandler})
	 * @throws ConnectException if an error occurred while attempting to connect to a remote address and port. 
	 * @throws IOException if an exception occurs
	 * @throws IllegalStateException if a request has already been forwarded 
	 */
	public void forward(IHttpRequest request, IHttpResponseHandler responseHandler) throws IOException, ConnectException, IllegalStateException;
	
	
	/**
	 * destroy the exchange and the underlying connection
	 */
	public void destroy();
}
