/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;


import java.net.URL;
import java.util.Enumeration;
import java.util.Map;
import java.util.Set;






/**
 * 
 * @author grro@xsocket.org
 */
public interface IHttpRequestHeader extends IHttpHeader {
	 
	
    
	/**
	 * Returns the fully qualified name of the client 
	 * or the last proxy that sent the request. If the 
	 * engine cannot or chooses not to resolve the hostname
	 * (to improve performance), this method returns the 
	 * dotted-string form of the IP address.
	 * 
	 * @return a String containing the fully qualified name of the client
	 */
	public String getRemoteHost();
	
	/**
	 * Returns the Internet Protocol (IP) source 
	 * port of the client or last proxy that sent the request.
	 * 
	 * @return an integer specifying the port number
	 */
	public int getRemotePort();
	

	/**
	 * Returns the Internet Protocol (IP) address of the client or last proxy that
	 * sent the request. For HTTP servlets, same as the value of the CGI variable REMOTE_ADDR
	 * 
	 * @return a String containing the IP address of the client that sent the request
	 */
	public String getRemoteAddr();
	
	
	/**
	 * Returns the name and version of the protocol the request 
	 * uses in the form protocol/majorVersion.minorVersion, for example, HTTP/1.1.  
	 * 
	 * @return a String containing the protocol name and version number
	 */
	public String getProtocol();
	
	
	
	/**
	 * Returns the query string that is contained in the request URL 
	 * after the path. This method returns <code>null</code> 
	 * if the URL does not have a query string.
	 * 
	 * @return a String containing the query string or null if the URL contains no query string. 
	 */
	public String getQueryString();
	
	
	
	/**
	 * Returns the part of this request's URL from the protocol
	 * name up to the query string in the first line of the HTTP request.
	 * 
	 * @return a String containing the part of the URL from the protocol 
	 *         name up to the query string
	 */
	public String getRequestURI();

	
	
	/**
	 * returns the full request url
	 * 
	 * @return the full request url
	 */
	public URL getTargetURL();
	
	
	/**
	 * set the target url
	 * 
	 * @param uri the target url
	 */
	public void setTargetURL(URL url);	
	
	
	/**
	 * updates the target url
	 * 
	 * @param host the host
	 * @param port the port
	 */
	public void updateTargetURL(String host, int port);
	
	
	/**
	 * Returns a boolean indicating whether this request 
	 * was made using a secure channel, such as HTTPS.
	 *   
	 * @return a boolean indicating if the request was made using a secure channel
	 */
	public boolean isSecure();
	
	
	
	/**
	 * Returns the name of the HTTP method with
	 * which this request was made, for example, GET, POST, or PUT.
	 * 
	 * @return a String specifying the name of the method 
	 */
	public String getMethod();
	
	
	
	/**
	 * Sets the name of the HTTP method 
	 * 
	 * @param method a String specifying the name of the method
	 */
	public void setMethod(String method);

	
	
	/**
	 * returns the scheme
	 * 
	 * @return the scheme
	 */
	public String getScheme();
	
	
	
	/**
	 * sets the scheme
	 *  
	 * @param scheme  the scheme 
	 */
	public void setScheme(String scheme);
	
	
	/**
	 * Returns the host name of the server to which the request was sent. 
	 * It is the value of the part before ":" in the Host  header value, 
	 * if any, or the resolved server name, or the server IP address.
	 * 
	 * @return the server name
	 */
	public String getServerName();
	
	
	/**
	 * Returns the port number to which the request was sent. It is the 
	 * value of the part after ":" in the Host  header value, if any, 
	 * or the server port where the client connection was accepted on.
	 * 
	 * @return the server port
	 */
	public int getServerPort();	
		
	
	
	/**
	 * sets a parameter 
	 * 
	 * @param parameterName   the parameter name 
	 * @param parameterValue  the parameter value
	 */
	public void setParameter(String parameterName, String parameterValue);


	
	/**
	 * returns the parameter map
	 * 
	 * @return the parameter map
	 */
	@SuppressWarnings("unchecked")
	public Map getParameterMap();
	
	
	/**
	 * Returns an Enumeration of String  objects containing the names 
	 * of the parameters contained in this request. If the request has no 
	 * parameters, the method returns an empty Enumeration. 
	 * 
	 * @return an Enumeration of String  objects, each String containing 
	 *         the name of a request parameter; or an empty Enumeration if the request has no parameters
	 */
	@SuppressWarnings("unchecked")
	public Enumeration getParameterNames();
	
	
	/**
	 * returns the parameter name set
	 *  
	 * @return the parameter name set
	 */
	public Set<String> getParameterNameSet();
	
	
	
	/**
	 * Returns the value of a request parameter as a String, or null 
	 * if the parameter does not exist. Request parameters are extra 
	 * information sent with the request. 
	 * 
	 * @param name a String specifying the name of the parameter
	 * @return a String representing the single value of the parameter 
	 */
	public String getParameter(String name);
	
	
	
	/**
	 * Get an Integer parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	public Integer getIntParameter(String name);
	
	
	
	/**
	 * Get an int parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	public int getIntParameter(String name, int defaultVal);
	
	
	
	/**
	 * Get an Long parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	public Long getLongParameter(String name);
	
	
	/**
	 * Get an long parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	public long getLongParameter(String name, long defaultVal);
	
	

	
	/**
	 * Get an Double parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	public Double getDoubleParameter(String name);
	
	
	
	/**
	 * Get an double parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	public double getDoubleParameter(String name, double defaultVal);
	
	
	
	
	/**
	 * Get an Float parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	public Float getFloatParameter(String name);
	
	
	
	/**
	 * Get an float parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	public float getFloatParameter(String name, float defaultVal);
	

	
	
	/**
	 * Get an Boolean parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	public Boolean getBooleanParameter(String name);
	
	
	
	/**
	 * Get an boolean parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	public boolean getBooleanParameter(String name, boolean defaultVal);

}

 