/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URLEncoder;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.logging.Level;
import java.util.logging.Logger;





/**
 * POST request
 * 
 * @author grro@xsocket.org
 */
public final class PostRequest extends HttpRequest  {
	
	private static final Logger LOG = Logger.getLogger(PostRequest.class.getName());
	
	
	
	/**
	 * constructor 
	 * 
	 * @param url   the url
	 */
	public PostRequest(String url) throws MalformedURLException {
		super(new HttpRequestHeader("POST", url));
	}
	
	/**
	 * constructor 
	 * 
	 * @param url           the url
	 * @param contentType   the content type
	 */
	public PostRequest(String url, String contentType) throws IOException, MalformedURLException {
		super("POST", url);
		setContentType(contentType);
	}

	
	


	/**
	 * constructor 
	 * 
	 * @param url          the url
	 * @param contentType  the content type 
	 * @param dataSource   the body 
	 * @throws IOException if an error occurs by reading the body 
	 */
	public PostRequest(String url, String contentType, FileChannel dataSource) throws IOException, MalformedURLException {
		super("POST", url, contentType, dataSource);
	}

	
	/**
	 * constructor 
	 * 
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 */
	public PostRequest(String url, String contentType, String body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);		
	}
	
	/**
	 * constructor 
	 * 
	 * @param url           the url
	 * @param encoding      the encoding
	 * @param contentType   the content type
	 * @param body          the body 
	 */
	public PostRequest(String url, String contentType, String encoding, String body) throws IOException, MalformedURLException {
		super("POST", url, contentType+ ";" + encoding, body);
		
	}

	
	/**
	 * constructor 
	 * 
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 */
	public PostRequest(String url, String contentType, byte[] body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);
	}
	
	

	
	/**
	 * constructor 
	 * 
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 */
	public PostRequest(String url, String contentType, ByteBuffer[] body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);
	}
	
	
	/**
	 * constructor
	 *  
	 * @param url             the url 
	 * @param formParameters  the FORM parameter
	 */
	public PostRequest(String url, String[] formParameters) throws IOException, MalformedURLException {
		super("POST", url, "application/x-www-form-urlencoded; charset=utf-8", createFormUrlEncodedBody(formParameters));
	}
	
	
	private static String createFormUrlEncodedBody(String[] formParameters) {
		StringBuilder sb = new StringBuilder();
		
		for (String param : formParameters) {
			try {
				String[] kvp = param.split("=");
				sb.append(URLEncoder.encode(kvp[0], "UTF-8") + "=" + URLEncoder.encode(kvp[1], "UTF-8") + "&");
			} catch (UnsupportedEncodingException usec) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by encoding param " + param + " " + usec.toString());
				}
			}
		}
		
		String body = sb.toString();
		if (body.length() > 0) {
			body = body.substring(0, body.length() - 1);
		}
		
		return body;
	}
}
