/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http.server;


import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.Execution;
import org.xsocket.ILifeCycle;
import org.xsocket.connection.http.HttpUtils;
import org.xsocket.connection.http.IBodyCompleteListener;
import org.xsocket.connection.http.IHttpConnectHandler;
import org.xsocket.connection.http.IHttpConnection;
import org.xsocket.connection.http.IHttpDisconnectHandler;
import org.xsocket.connection.http.IHttpExchange;
import org.xsocket.connection.http.IHttpHandler;
import org.xsocket.connection.http.IHttpRequest;
import org.xsocket.connection.http.IHttpRequestHandler;
import org.xsocket.connection.http.IHttpRequestTimeoutHandler;




/**
*
* @author grro@xsocket.org
*/

@Execution(Execution.NONTHREADED)
public final class Context implements IHttpRequestHandler, IHttpRequestTimeoutHandler, IHttpConnectHandler, IHttpDisconnectHandler, ILifeCycle, Cloneable  {

	private static final Logger LOG = Logger.getLogger(Context.class.getName());
	
	
	private final List<IHttpHandler> handlers = new ArrayList<IHttpHandler>();
	
	private final List<ILifeCycle> lifeCycleChain = new ArrayList<ILifeCycle>();
	
	private final List<IHolder> holders = new ArrayList<IHolder>();
	private final Map<String, IHolder> holderCache = HttpUtils.newMapCache(40);
	
	private boolean isOnConnectPathMultithreaded = false;
	private final List<IHttpConnectHandler> connectHandlerChain = new ArrayList<IHttpConnectHandler>();

	private boolean isOnRequestTimeoutPathMultithreaded = false;
	private final List<IHttpRequestTimeoutHandler> requestTimeoutHandlerChain = new ArrayList<IHttpRequestTimeoutHandler>();

	private boolean isOnDisconnectPathMultithreaded = false;
	private final List<IHttpDisconnectHandler> disconnectHandlerChain = new ArrayList<IHttpDisconnectHandler>();
	
	private String contextPath = null;
	
	
	/**
	 * constructor
	 * 
	 * @param contextPath   the context path
	 */
	public Context(String contextPath) {
		this.contextPath = contextPath;
	}

	
	/**
	 * constructor
	 * 
	 * @param contextPath   the context path
	 */
	public Context(String contextPath, Map<String, IHttpRequestHandler> handlers) {
		this.contextPath = contextPath;
		
		for (Entry<String, IHttpRequestHandler> entry : handlers.entrySet()) {
			addHandler(entry.getKey(), entry.getValue());
		}
	}

	
	/**
	 * constructor
	 * 
	 * @param parentContext   the parent context
	 * @param contextPath     the context path
	 */
	public Context(Context parentContext, String contextPath) {
		this.contextPath = contextPath;
		
		parentContext.addContext(this);
	}
	
	
	private void addContext(Context ctx) {
		holders.add(new ContextHolder(ctx, contextPath)); 
		sortHolderList();
	}
	
	
	
	/**
	 * adds a handler to the current context 
	 * 
	 * @param pattern          the pattern
	 * @param requestHandler   the handler (supported: {@link IHttpRequestHandler}, {@link ILifeCycle})
	 */
	public void addHandler(String pattern, IHttpRequestHandler requestHandler) {
		
		holders.add(new RequestHandlerHolder(pattern, requestHandler, HttpUtils.isRequestHandlerMultithreaded(requestHandler), HttpUtils.isRequestHandlerInvokedOnMessageReceived(requestHandler)));
		sortHolderList();
		
		handlers.add(requestHandler);		
		computePath();
	}
	
	
	private void computePath() {
		
		lifeCycleChain.clear();
		
		connectHandlerChain.clear();
		isOnConnectPathMultithreaded = false;
	
		requestTimeoutHandlerChain.clear();
		isOnRequestTimeoutPathMultithreaded = false;
		
		disconnectHandlerChain.clear();
		isOnDisconnectPathMultithreaded = false;
	
		
		for (IHttpHandler handler : handlers) {

			if (ILifeCycle.class.isAssignableFrom(handler.getClass())) {
				lifeCycleChain.add((ILifeCycle) handler);
			}
			
			if (HttpUtils.isConnectHandler(handler)) {
				connectHandlerChain.add((IHttpConnectHandler) handler);
				isOnConnectPathMultithreaded = isOnConnectPathMultithreaded || HttpUtils.isConnectHandlerMultithreaded(handler);
			}

			
			if (HttpUtils.isDisconnectHandler(handler)) {
				disconnectHandlerChain.add((IHttpDisconnectHandler) handler);
				isOnDisconnectPathMultithreaded = isOnDisconnectPathMultithreaded || HttpUtils.isDisconnectHandlerMultithreaded(handler);
			}
			
			if (HttpUtils.isRequestTimeoutHandler(handler)) {
				requestTimeoutHandlerChain.add((IHttpRequestTimeoutHandler) handler);
				isOnRequestTimeoutPathMultithreaded = isOnRequestTimeoutPathMultithreaded || HttpUtils.isRequestTimeoutHandlerMultithreaded(handler);
			}
		}		
	}
	 
	
	
	private void sortHolderList() {
		Comparator<IHolder> comparator = new Comparator<IHolder>() {

			public int compare(IHolder o1, IHolder o2) {
				return (0 - o1.getPattern().compareTo(o2.getPattern()));
			}
		};

		Collections.sort(holders, comparator);
	}
	

	/**
	 * {@inheritDoc}
	 */
	public void onInit() {
		for (IHolder holder : holders) {
			holder.onInit();
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public void onDestroy() throws IOException {
		for (IHolder holder : holders) {
			holder.onDestroy();
		}		
	}
	
	String getContextPath() {
		return contextPath;
	}

	
	List<String> getMapping() {
		List<String> result = new ArrayList<String>();
		
		for (IHolder holder : holders) {
			result.add("[" + holder.getPattern() + "] -> " + holder.getTarget());
		}
		
		Collections.sort(result);
		
		return result;
	}
	
	List<IHttpRequestHandler> getHandlers() {
		List<IHttpRequestHandler> result = new ArrayList<IHttpRequestHandler>();
		
		for (IHolder holder : holders) {
			result.add(holder.getTarget());
		}
		
		return result;
	}
	

	
	public boolean onConnect(IHttpConnection httpConnection) throws IOException {
	
		if (connectHandlerChain.isEmpty()) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer("no connect handler set. ignore callback");
			}
			
			return false;
		}
		
		
		if (isOnConnectPathMultithreaded) {
			((HttpServerConnection) httpConnection).processMultithreaded(new OnConnectCaller(httpConnection));
			
		} else {
			((HttpServerConnection) httpConnection).processNonthreaded(new OnConnectCaller(httpConnection));
		}
		
		return true;
	}
	

	private boolean callOnConnectCallback(IHttpConnection connection) throws IOException {
		
		boolean handled = false;
		for (IHttpConnectHandler connectHandler : connectHandlerChain) {
			boolean result = connectHandler.onConnect(connection);
			handled = handled || result;
		}	
		
		return handled;
	}	
	
	
	
	
	
	public boolean onDisconnect(IHttpConnection httpConnection) throws IOException {
		
		if (disconnectHandlerChain.isEmpty()) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer("no disconnect handler set. ignore callback");
			}
			
			return false;
		}
		
		
		
		if (isOnDisconnectPathMultithreaded) {
			((HttpServerConnection) httpConnection).processMultithreaded(new OnDisconnectCaller(httpConnection));
			
		} else {
			((HttpServerConnection) httpConnection).processNonthreaded(new OnDisconnectCaller(httpConnection));
		}
		
		return true;
	}
	

	
	private boolean callOnDisconnectCallback(IHttpConnection connection) throws IOException {
		
		boolean handled = false;
		
		for (IHttpDisconnectHandler disconnectHandler : disconnectHandlerChain) {
			boolean result = disconnectHandler.onDisconnect(connection);
			handled = handled || result;
		}	
		
		return handled;
	}
	

	
	
	/**
	 * {@inheritDoc}
	 */
	public void onRequest(IHttpExchange exchange) throws IOException {
		
		String path = exchange.getRequest().getRequestURI();
		
		if (path.startsWith(contextPath)) {
			path = path.substring(contextPath.length(), path.length());
			
			onRequest(path, exchange, contextPath);
			
		} else {
			exchange.sendError(404);
		}
	}	
	



	private void onRequest(String path, IHttpExchange exchange, String totalContextPath) throws IOException {
		
		if (holderCache.containsKey(exchange.getRequest())){
			holderCache.get(exchange.getRequest()).onRequest(path, exchange, totalContextPath);
			return;
		}
			
		for (IHolder holder : holders) {
			if (holder.match(path)) {
				holderCache.put(path, holder);
				holder.onRequest(path, exchange, totalContextPath);
				return;
			}
		}
		
		exchange.sendError(404);
	}	
	
		
	public boolean onRequestTimeout(IHttpConnection connection) throws IOException {

		if (requestTimeoutHandlerChain.isEmpty()) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer("no request timeout handler set. ignore callback");
			}
			
			return false;
		}
		
		HttpServerConnection httpConnection = (HttpServerConnection) connection; 

		
		if (isOnRequestTimeoutPathMultithreaded) {
			httpConnection.processMultithreaded(new OnRequestTimeoutCaller(httpConnection));
			
		} else {
			httpConnection.processNonthreaded(new OnRequestTimeoutCaller(httpConnection));
		}
		
		return true;
	}
	

	
	private boolean callOnRequestTimeoutCallback(IHttpConnection connection) throws IOException {
		
		for (IHttpRequestTimeoutHandler requestTimeoutHandler : requestTimeoutHandlerChain) {
			boolean result = requestTimeoutHandler.onRequestTimeout(connection);
			if (result == true) {
				return true;
			}
		}	
		
		return false;
	}
		
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected Object clone() throws CloneNotSupportedException {
		Context copy = (Context) super.clone();
		return copy;
	}
	

	/**
	 * copies the context
	 * @return the copy
	 */
	Context copy() {
		try {
			return (Context) clone();
		} catch (CloneNotSupportedException cnse) {
			throw new RuntimeException(cnse.toString());
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder("\"" + contextPath + "\"->{");
		for (IHolder holder  : holders) {
			sb.append(holder + " ");
		}
		return sb.toString().trim() + "}";
	}
	
	
	private interface IHolder {
				
		public void onInit();
		
		public void onDestroy() throws IOException;
		
		public void onRequest(String path, IHttpExchange exchange, String totalContextPath) throws IOException;
		
		public boolean match(String requestedRessource);
		
		public String getPattern();
		
		public IHttpRequestHandler getTarget();
	}
	
	
	private static final class ContextHolder implements IHolder {

		private Context context = null;
		
		ContextHolder(Context context, String parentContext) {
			this.context = context.copy();
		}
		
		public void onInit() {
			context.onInit();
		}
		
		public void onDestroy() throws IOException {
			context.onDestroy();
		}
		
		
		@Execution(Execution.NONTHREADED)
		public void onRequest(String path, IHttpExchange exchange, String totalContextPath) throws IOException {
			path = path.substring(context.contextPath.length(), path.length());
			context.onRequest(path, exchange, totalContextPath + context.contextPath);
		}
		
		public boolean match(String requestedRessource) {
			return requestedRessource.startsWith(context.contextPath) || requestedRessource.equals(context.contextPath);
		}
		
		public String getPattern() {
			return context.contextPath;
		}
		
		public IHttpRequestHandler getTarget() {
			return context;
		}
				
		@Override
		public String toString() {
			return context.toString();
		}
	}	

	
	

	private final class OnConnectCaller implements Runnable {
		
		private IHttpConnection httpConnection = null;
		
		public OnConnectCaller(IHttpConnection httpConnection) {
			this.httpConnection = httpConnection;
		}
		
		public void run() {
			try {
				callOnConnectCallback(httpConnection);
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("Error occured by calling onConnect callback " + ioe.toString());
				}
			}
		}
	}
	
	private final class OnDisconnectCaller implements Runnable {
		
		private IHttpConnection httpConnection = null;
		
		public OnDisconnectCaller(IHttpConnection httpConnection) {
			this.httpConnection = httpConnection;
		}
		
		public void run() {
			try {
				callOnDisconnectCallback(httpConnection);
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("Error occured by calling onDisconnect callback " + ioe.toString());
				}
			}
		}
	}
	
	
	private final class OnRequestTimeoutCaller implements Runnable {
	
		private IHttpConnection httpConnection = null;
		
		public OnRequestTimeoutCaller(IHttpConnection httpConnection) {
			this.httpConnection = httpConnection;
		}
		
		public void run() {
			try {
				callOnRequestTimeoutCallback(httpConnection);
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("Error occured by calling onRequestTimeout callback " + ioe.toString());
				}
			}
		}
	}
	

	
	
	private static final class RequestHandlerHolder implements IHolder {

		private String path = null;
		private String pattern = null; 
		private boolean isWildcardPath = false;
		private boolean isWildcardPathExt = false;
		
		private IHttpRequestHandler handler = null; 
		private boolean isMultithreaded = true;
		private boolean invokeOnMessageReceived = false;
		private boolean isLifeCycle = false;

			 
		RequestHandlerHolder(String pattern, IHttpRequestHandler handler, boolean isMultithreaded, boolean invokeOnMessageReceived) {
			this.handler = handler;
			this.pattern = pattern;
			this.path = pattern;
			
			this.isMultithreaded = isMultithreaded;
			this.invokeOnMessageReceived = invokeOnMessageReceived;
			
			isLifeCycle = (handler instanceof ILifeCycle);
			
			if (pattern.endsWith("/*")) {
				isWildcardPath = true;
				path = pattern.substring(0, pattern.indexOf("/*"));
					
			} else if (pattern.startsWith("*")) {
				path = pattern.substring(1, pattern.length());
				isWildcardPathExt = true;

			} else {
				isWildcardPath = false;
				path = pattern;
			}
		}
		
		
		public void onInit() {
			if (isLifeCycle) {
				((ILifeCycle) handler).onInit();
			}
		}
		
		public void onDestroy() throws IOException {
			if (isLifeCycle) {
				((ILifeCycle) handler).onDestroy();
			}
		}
		
		
		@Execution(Execution.NONTHREADED)
		public void onRequest(String reqPath, final IHttpExchange exchange, String totalContextPath) throws IOException {
			
			IHttpRequest request = exchange.getRequest();
			request.setContextPath(totalContextPath);
			request.setRequestHandlerPath(path);

			
			// mode MESSAGE_RECEIVED			
			if (invokeOnMessageReceived) {
				if (request != null) {
					if (request.hasBody()) {
						
						IBodyCompleteListener completeListener = new IBodyCompleteListener() {
							
							@Execution(Execution.NONTHREADED)
							public final void onComplete() {
								// call handlers call back method if is HEADER_MESSAGE triggered
								if (invokeOnMessageReceived)  {
									performOnRequest(exchange);
								}
							}
						};
						request.getNonBlockingBody().addCompleteListener(completeListener);

					} else {
						performOnRequest(exchange);
					}
				}
				
			// mode HEADER_RECEIVED			
			} else {
				performOnRequest(exchange);
			}
		}
		
		
		private void performOnRequest(final IHttpExchange exchange) {
			
			Runnable task = new Runnable() {
				public void run() {
					try {
						handler.onRequest(exchange);
					} catch (Exception e) {
						if (LOG.isLoggable(Level.FINE)) {
							LOG.fine("error occured by calling on request " + handler + " " + e.toString());
						}
					}
				}
			}; 
			
			
			if (isMultithreaded) {
				((HttpServerConnection) exchange.getConnection()).processMultithreaded(task);
				
			} else {
				((HttpServerConnection) exchange.getConnection()).processNonthreaded(task);
			}
		}

			
		public boolean match(String requestedRessource) {
				
			if (isWildcardPath) {
				return requestedRessource.startsWith(path) || requestedRessource.equals(path);
					
			} else if (isWildcardPathExt) {
				return requestedRessource.endsWith(path);
					
			} else {
				return requestedRessource.equals(path);
			}
		}
		
		
		public String getPattern() {
			return pattern;
		}
		
		public IHttpRequestHandler getTarget() {
			return handler;
		}
		
		@Override
		public String toString() {
			return "\"" + pattern + "\"->" + handler.getClass().getSimpleName();
		}
	}	
}