/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http.server;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.channels.FileChannel;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.Execution;
import org.xsocket.connection.http.BodyDataSink;
import org.xsocket.connection.http.HttpResponse;
import org.xsocket.connection.http.HttpResponseHeader;
import org.xsocket.connection.http.HttpUtils;
import org.xsocket.connection.http.IHttpExchange;
import org.xsocket.connection.http.IHttpRequest;
import org.xsocket.connection.http.IHttpRequestHandler;
import org.xsocket.connection.http.InvokeOn;




/**
 * File service handler which returns the requested file
 *
 * @author grro@xsocket.org
 */
public final class FileServiceHandler implements IHttpRequestHandler {
	
	private static final Logger LOG = Logger.getLogger(FileServiceHandler.class.getName());

	public static final boolean SHOW_DIRECTORY_TREE_DEFAULT = false;
	
	private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");
	private static final Map<String, String> MIME_TYPE_MAPPING = HttpUtils.getMimeTypeMapping();

	private String fileBasepath = null;
	private boolean isShowDirectoryTree = false;
	
	
	/**
	 * constructor
	 * 
	 * @param fileBasepath   the base path
	 * 
	 * @throws FileNotFoundException  if the base path not exists
	 */
	public FileServiceHandler(String fileBasepath) throws FileNotFoundException {
		this(fileBasepath, SHOW_DIRECTORY_TREE_DEFAULT);
	}
	
	/**
	 * constructor
	 * 
	 * @param fileBasepath         the base path
	 * @param isShowDirectoryTree  true, if the directory tree will been shown, if the requests file is a directory
	 * 
	 * @throws FileNotFoundException  if the base path not exists
	 */
	public FileServiceHandler(String fileBasepath, boolean isShowDirectoryTree) throws FileNotFoundException {
		this.fileBasepath = fileBasepath;
		this.isShowDirectoryTree = isShowDirectoryTree;
		
		if (!new File(fileBasepath).exists()) {
			throw new FileNotFoundException("base path "+ fileBasepath + "does not exits");
		}
	}
	
	
	boolean isShowDirectoryTree() {
		return isShowDirectoryTree;
	}
	
	void setShowDirectoryTree(boolean isShowDirectoryTree) {
		this.isShowDirectoryTree = isShowDirectoryTree;
	}
	
	String getBasepath() {
		return fileBasepath;
	}
	
	void setBasepath(String fileBasepath) {
		this.fileBasepath = fileBasepath;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Execution(Execution.MULTITHREADED)
	@InvokeOn(InvokeOn.MESSAGE_RECEIVED)
	public void onRequest(IHttpExchange exchange) throws IOException {
		
		IHttpRequest request = exchange.getRequest();
		
		// only GET or POST is supported by this handler
		if (request.getMethod().equalsIgnoreCase("GET") || request.getMethod().equalsIgnoreCase("POST")) {
			
			String requestURI = URLDecoder.decode(request.getRequestURI(), "UTF-8");
			int ctxLength = request.getContextPath().length() + request.getRequestHandlerPath().length();
			
			
			if (requestURI.length() > ctxLength) {
				String filepath = requestURI.substring(ctxLength, requestURI.length());
		
				if (filepath.length() > 0) {
	
					String path = fileBasepath + filepath;
					
					path = path.replace("/", "\\");
					if (path.endsWith("\\")) {
						path = path.substring(0, path.length() - 1);
					}
	
					File file = new File(path);
					if (file.exists()) {
						
						if (file.isFile()) {
							if (LOG.isLoggable(Level.FINE)) {
								LOG.fine(filepath + " requested. returning data");
							}
							
							HttpResponseHeader responseHeader = new HttpResponseHeader(200);
							
							int pos = filepath.lastIndexOf(".");
							if (pos != -1) {
								String extension = filepath.substring(pos + 1, filepath.length());
								String mimeType = MIME_TYPE_MAPPING.get(extension);
								if (mimeType != null) {
									responseHeader.setContentType(mimeType);
								}
							}
		
							
							FileChannel fc = new RandomAccessFile(file, "r").getChannel();
														
							BodyDataSink outChannel = exchange.send(responseHeader, (int) fc.size());
							outChannel.transferFrom(fc);
							
							outChannel.close();
							fc.close();
							
							return;
							
						} else {
							if (isShowDirectoryTree) {
								if (file.isDirectory()) {
									String body = printDirectoryTree(request, file);
									exchange.send(new HttpResponse(200, "text/html", body));
									return;
								}
							} 
						}
						
					} else {
						
						if (isShowDirectoryTree) {
							if (file.isDirectory()) {
								String body = printDirectoryTree(request, file);
								exchange.send(new HttpResponse(200, "text/html", body));
								return;
							}
						} 
					}
				} 
			} else {
				exchange.sendError(404, request.getRequestURI() + " not found");
			}
		}		
		
		
		if (isShowDirectoryTree) {
			String body = printDirectoryTree(request, new File(fileBasepath));
			exchange.send(new HttpResponse(200, "text/html", body));		
			return;
		}
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("unsupported method " + request.getMethod() + " returning 404");
		}					
		
		exchange.sendError(404, request.getRequestURI() + " not found");		
	}	
	
	
	private String printDirectoryTree(IHttpRequest request, File directory) throws IOException {
		StringBuilder sb = new StringBuilder();
		
		String dir = URLDecoder.decode(request.getRequestURI(), "UTF-8"); 
		
		sb.append("<html>\r\n");
		sb.append("  <!-- This page is auto-generated by xSocket-http (http://xsocket.org) -->\r\n");
		sb.append("  <head>\r\n");
		sb.append("    <title>Index of " + dir + "</title>\r\n");
		sb.append("  </head>\r\n");
		sb.append("  <body>\r\n");
		sb.append("    <H1 style=\"color:#0a328c;font-size:1.5em;\">Index of " + dir + "</H1>\r\n");
		sb.append("    <p style=\"font-size:0.8em;\">\r\n");

		
		sb.append("    <table border=\"0\" style=\"color:#0a328c;font-size:1.0em;\">\r\n");


		
		for (File file : directory.listFiles()) {
			sb.append("      <tr>");
			
			sb.append("        <td align=\"right\">");
			
			if (file.isDirectory()) {
				sb.append("[DIR]");
			} else {
				sb.append("[TXT]");
			}
			
			sb.append("        </td>\r\n");

			
			sb.append("        <td>");
			
			String path = request.getRequestURI().substring(request.getRequestHandlerPath().length() + 1, request.getRequestURI().length());
			
			if (path.length() > 0) {
				String[] parts = path.split("/");
				path = parts[parts.length -1];

				sb.append("<a href=" + path + "/");
			} else {
				sb.append("<a href=");

			}
			
			sb.append(URLEncoder.encode(file.getName(), "UTF-8") + "> " + file.getName() + "</a>");
			sb.append("        </td>\r\n");
			
			sb.append("        <td>");
			sb.append(DataConverter.toFormatedDate(file.lastModified()));
			sb.append("        </td>\r\n");
			
			sb.append("        <td align=\"right\">");
			if (!file.isDirectory()) {
				sb.append(DataConverter.toFormatedBytesSize(file.length()));
			} else {
				sb.append("-");
			}
			sb.append("        </td>\r\n");
			
			sb.append("      </tr>");
		}
		
		sb.append("    </table>\r\n");
		
		sb.append("    </p>\r\n");
		sb.append("    <p style=\"font-size:0.8em;\">" + DATE_FORMAT.format(new Date()) + "    xSocket-http (" + 
				  HttpUtils.getVersionInfo() + ") at " + request.getServerName() + 
				  " Port " + request.getServerPort() + "</p>\r\n");
		sb.append("  </body>\r\n");
		sb.append("</html>\r\n");
		return sb.toString(); 
	}
}
