/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http.server;


import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.xsocket.Execution;
import org.xsocket.ILifeCycle;
import org.xsocket.Resource;
import org.xsocket.connection.IConnectHandler;
import org.xsocket.connection.IConnectionScoped;
import org.xsocket.connection.INonBlockingConnection;
import org.xsocket.connection.IServer;
import org.xsocket.connection.http.HttpUtils;
import org.xsocket.connection.http.IHttpConnectHandler;
import org.xsocket.connection.http.IHttpDisconnectHandler;
import org.xsocket.connection.http.IHttpHandler;
import org.xsocket.connection.http.IHttpRequestHandler;
import org.xsocket.connection.http.IHttpRequestTimeoutHandler;
import org.xsocket.connection.http.client.HttpClient;



/**
 * The http protocol adapter. The protocol adapter implements the {@link IConnectHandler} interface, and
 * "maps" each incoming connection into a {@link HttpServerConnection}. <br> 
 * HttpProtcol adapter is required, if the {@link HttpServer} will not be used.
 * 
 * Example:
 * <pre>
 * 
 * // establishing a http server based on a mutliplexed tcp connection
 * // for multiplexex connections see http://xsocket.sourceforge.net/multiplexed/tutorial/V2/TutorialMultiplexed.htm    
 *    
 * IRequestHandler hdl = new MyRequestHandler();
 * IConnectHandler httpAdapter = new HttpProtocolAdapter(hdl);
 * 
 * 
 * IServer mutliplexedHttpServer = new Server(port, new MultiplexedProtocolAdapter(httpAdapter)));
 * ConnectionUtils.start(mutliplexedHttpServer);
 * ...
 * </pre> 
 * 
 * 
 * @author grro
 */
@Execution(Execution.NONTHREADED)
public final class HttpProtocolAdapter implements IConnectHandler, ILifeCycle, MBeanRegistration {
	
	private static final Logger LOG = Logger.getLogger(HttpProtocolAdapter.class.getName());
	

	
	@Resource
	private IServer server = null;
	
	
	private IHttpHandler handler = null;
	private boolean isLifeCylceHandler = false;
	private boolean isConnectionScoped = false;
	private boolean isConnectHandler = false;
	private boolean isConnectHandlerMultithreaded = false;
	private boolean isDisconnectHandler = false;
	private boolean isDisconnectHandlerMultithreaded = false;
	private boolean isRequestTimeoutHandler = false;
	private boolean isRequestTimeoutHandlerMultithreaded = false;
	private boolean isRequestHandler = false;
	private boolean isRequestHandlerMultithreaded = false;
	private boolean isRequestHandlerInvokeOnMessageReceived = false;
	
	private HttpClient httpClient = null;

	private Integer requestTimeoutSec = null;
	
	private boolean isCloseOnSendingError = false;
	
	
	// keep alive support 
	private Integer maxTransactions = null;
	
	
	
	/**
	 * constructor 
	 * 
	 * @param handler the handler (supported: {@link IHttpRequestHandler}, {@link IHttpRequestTimeoutHandler}, {@link IHttpConnectHandler}, {@link IHttpDisconnectHandler}, {@link ILifeCycle})
	 */
	public HttpProtocolAdapter(IHttpHandler handler) {
		this.handler = handler;
		
		isLifeCylceHandler = (ILifeCycle.class.isAssignableFrom(handler.getClass()));
		isConnectionScoped = (IConnectionScoped.class.isAssignableFrom(handler.getClass()));
				
		isConnectHandler = HttpUtils.isConnectHandler(handler);
		isConnectHandlerMultithreaded = HttpUtils.isConnectHandlerMultithreaded(handler);
		isDisconnectHandler = HttpUtils.isDisconnectHandler(handler);
		isDisconnectHandlerMultithreaded = HttpUtils.isDisconnectHandlerMultithreaded(handler);
		isRequestTimeoutHandler = HttpUtils.isRequestTimeoutHandler(handler);
		isRequestTimeoutHandlerMultithreaded = HttpUtils.isRequestTimeoutHandlerMultithreaded(handler);
		isRequestHandler = HttpUtils.isRequestHandler(handler);
		isRequestHandlerMultithreaded = HttpUtils.isRequestHandlerMultithreaded(handler);
		isRequestHandlerInvokeOnMessageReceived = HttpUtils.isRequestHandlerInvokedOnMessageReceived(handler);
	}
	
	
	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception {
		return ServerUtils.exportMbean(server, name, handler);
	}
	
	public void postRegister(Boolean registrationDone) {
	}
	
	public void preDeregister() throws Exception {
	}
	
	public void postDeregister() {
	}
	

	
	public void setForwardHttpClient(HttpClient httpClient) {
		this.httpClient = httpClient;
	}
	
	/**
	 * sets the message receive timeout
	 * 
	 * @param request the message receive timeout
	 */
	public void setRequestTimeoutSec(int requesttimeout) {
		this.requestTimeoutSec = requesttimeout;
	}

	
	/**
	 * returns the message request timeout
	 * 
	 * @return the message request timeout
	 */
	public int getRequestTimeoutSec() {
		return requestTimeoutSec;
	}	
	
	
	/**
	 * set is if the server-side connection will closed, if an error message is sent
	 * 
	 * @param isCloseOnSendingError if the connection will closed, if an error message is sent
	 */
	public void setCloseOnSendingError(boolean isCloseOnSendingError) {
		this.isCloseOnSendingError = isCloseOnSendingError;
	}
	
	
	/**
	 * returns if the server-side connection will closed, if an error message will be sent
	 * @return true, if the connection will closed by sending an error message 
	 */
	public boolean isCloseOnSendingError() {
		return isCloseOnSendingError;
	}
	
	
	/**
	 * set the max transactions per connection. Setting this filed causes that 
	 * a keep-alive response header will be added 
	 * 
	 * @param maxTransactions  the max transactions 
	 */
	public void setMaxTransactions(int maxTransactions) {
		this.maxTransactions = maxTransactions;
	}
	
	

	public String getVersion() {
		return HttpUtils.getVersionInfo();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void onInit() {
		server.setStartUpLogMessage(server.getStartUpLogMessage() + "; http " + HttpUtils.getVersionInfo());

		ServerUtils.injectServerField(handler, server);
		ServerUtils.injectProtocolAdapter(handler, this);
		
		if (isLifeCylceHandler) {
			((ILifeCycle) handler).onInit();
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void onDestroy() {
		if (isLifeCylceHandler) {
			try {
				((ILifeCycle) handler).onDestroy();
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("exception occured by destroying " + handler + " " + ioe.toString());
				}
			}
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public boolean onConnect(INonBlockingConnection connection) throws IOException {
		
		HttpServerConnection httpCon = null;
		
		if (isConnectionScoped) {
			try {
				IHttpRequestHandler handlerCopy = (IHttpRequestHandler) ((IConnectionScoped) handler).clone();
				httpCon = new HttpServerConnection(connection, isConnectHandler, isConnectHandlerMultithreaded, isDisconnectHandler, isDisconnectHandlerMultithreaded, isRequestTimeoutHandler, isRequestTimeoutHandlerMultithreaded, isRequestHandler, isRequestHandlerMultithreaded, isRequestHandlerInvokeOnMessageReceived, handlerCopy, isCloseOnSendingError, httpClient);
			} catch (CloneNotSupportedException clne) {
				throw new RuntimeException("couldn't clone handler " + handler + " " + clne.toString());
			}
			
		} else {
			httpCon = new HttpServerConnection(connection, isConnectHandler, isConnectHandlerMultithreaded, isDisconnectHandler, isDisconnectHandlerMultithreaded, isRequestTimeoutHandler, isRequestTimeoutHandlerMultithreaded, isRequestHandler, isRequestHandlerMultithreaded, isRequestHandlerInvokeOnMessageReceived, handler, isCloseOnSendingError, httpClient);
		}
		
		if (maxTransactions != null) {
			httpCon.setMaxTransactions(maxTransactions);
		}
		
		if (requestTimeoutSec != null) {
			httpCon.setRequestTimeoutSec(requestTimeoutSec);
		}
		
		return true;
	}	
}