/*
 *  Copyright (c) xsocket.org, 2006 - 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xsocket.org/
 */
package org.xsocket.connection.http.server;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.ref.WeakReference;
import java.net.ConnectException;
import java.net.SocketTimeoutException;
import java.nio.BufferUnderflowException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimerTask;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;



import org.xsocket.DataConverter;
import org.xsocket.Execution;
import org.xsocket.ILifeCycle;
import org.xsocket.connection.INonBlockingConnection;
import org.xsocket.connection.http.AbstractHttpConnection;
import org.xsocket.connection.http.BodyDataSink;
import org.xsocket.connection.http.HttpRequest;
import org.xsocket.connection.http.HttpRequestHeader;
import org.xsocket.connection.http.HttpResponse;
import org.xsocket.connection.http.HttpResponseHeader;
import org.xsocket.connection.http.HttpUtils;
import org.xsocket.connection.http.IBodyCloseListener;
import org.xsocket.connection.http.IBodyCompleteListener;
import org.xsocket.connection.http.IHttpConnectHandler;
import org.xsocket.connection.http.IHttpConnection;
import org.xsocket.connection.http.IHttpDisconnectHandler;
import org.xsocket.connection.http.IHttpExchange;
import org.xsocket.connection.http.IHttpHandler;
import org.xsocket.connection.http.IHttpRequest;
import org.xsocket.connection.http.IHttpRequestHandler;
import org.xsocket.connection.http.IHttpRequestHeader;
import org.xsocket.connection.http.IHttpRequestTimeoutHandler;
import org.xsocket.connection.http.IHttpResponse;
import org.xsocket.connection.http.IHttpResponseHandler;
import org.xsocket.connection.http.IHttpResponseHeader;
import org.xsocket.connection.http.IHttpResponseTimeoutHandler;
import org.xsocket.connection.http.NonBlockingBodyDataSource;
import org.xsocket.connection.http.client.HttpClient;




/**
 * Represents the server side endpoint implementation of a http connection. Typically a
 * HttpServerConnection will be created by the {@link HttpProtocolAdapter}. A HttpServerConnection
 * can also be created manually by passing over a {@link INonBlockingConnection} and the
 * server handler.
 *
 * @author grro@xsocket.org
 */
public class HttpServerConnection extends AbstractHttpConnection implements IHttpConnection {

	private static final Logger LOG = Logger.getLogger(HttpServerConnection.class.getName());

	public static final boolean DEFAULT_REMOVE_REQUEST_CONNECTION_HEADER = false;
	public static final Integer DEFAULT_RECEIVE_TIMEOUT_MILLIS = Integer.MAX_VALUE; 
	public static final boolean DEFAULT_AUTOHANDLE_EXPECT_100CONTINUE_HEADER = true;


	
	private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");

	
	public static final String SHOW_DETAILED_ERROR_KEY = "org.xsocket.connection.http.server.showDetailedError";
	public static final String SHOW_DETAILED_ERROR_DEFAULT = "false";
	private static boolean showDetailedError;
	
	static {
		showDetailedError = Boolean.parseBoolean(System.getProperty(SHOW_DETAILED_ERROR_KEY, SHOW_DETAILED_ERROR_DEFAULT));
	}
	
	
	// http client
	private static HttpClient defaultHttpClient = null;
	private HttpClient httpClient = null;
	

	
	// handler support
	private Object handler = null;
	private ConnectHandlerAdapter connectHandlerAdapter = null;
	private DisconnectHandlerAdapter disconnectHandlerAdapter = null;
	private RequestTimeoutHandlerAdapter requestTimeoutHandlerAdapter = null;
	private boolean isRequestHandler = false;
	private boolean isRequestHandlerMultithreaded = true;
	private boolean isRequestHandlerInvokeOnMessageReceived = false;
	private final OnRequestCaller onRequestCaller = new OnRequestCaller();
	private final BodyCompleteListener bodyCompleteListener = new BodyCompleteListener();
	
	// close handling
	private boolean isCloseAfterResponse = false;


	// request message management
	private final List<IHttpRequest> requestQueue = new ArrayList<IHttpRequest>();
	private int requestQueueVersion = 0;
	private final AtomicInteger openRequests = new AtomicInteger(0);
	

	// auto handle
	boolean isAutohandleExpect100ContinueHeader = DEFAULT_AUTOHANDLE_EXPECT_100CONTINUE_HEADER;


	// timeout support
	private static final long MIN_WATCHDOG_PERIOD_MILLIS = 10 * 1000;
	private TimeoutWatchDogTask watchDogTask = null;
	private long lastTimeMessageReceivedMillis = System.currentTimeMillis();
	private Integer requestTimeoutSec = null;

	
	
	// max transaction support
	private Integer maxTransactions = null;
	

	private final AtomicInteger countReceivedMessages = new AtomicInteger(0);
	private final AtomicInteger countSendMessages = new AtomicInteger(0);




	/**
	 * constructor
	 *
	 * @param connection      the underlying tcp connection
	 * @param handler         the server handler (supported: {@link IHttpRequestHandler}, {@link IHttpConnectHandler}, {@link IHttpDisconnectHandler}, {@link IHttpConnectionTimeoutHandler}, {@link IHttpResponseTimeoutHandler}, {@link ILifeCycle})
	 * @throws IOException if an exception occurs
	 */
	public HttpServerConnection(INonBlockingConnection connection, IHttpHandler handler) throws IOException {
		this(connection, HttpUtils.isConnectHandler(handler), HttpUtils.isConnectHandlerMultithreaded(handler), HttpUtils.isDisconnectHandler(handler), HttpUtils.isDisconnectHandlerMultithreaded(handler), HttpUtils.isRequestTimeoutHandler(handler), HttpUtils.isRequestTimeoutHandlerMultithreaded(handler), HttpUtils.isRequestHandler(handler), HttpUtils.isRequestHandlerMultithreaded(handler), HttpUtils.isRequestHandlerInvokedOnMessageReceived(handler), handler, false, null);
	}


	HttpServerConnection(INonBlockingConnection connection, boolean isConnectHandler, boolean isConnectHandlerMultithreaded, boolean isDisconnectHandler, boolean isDisconnectHandlerMultithreaded, boolean isRequestTimeoutHandler, boolean isRequestTimeoutHandlerMultithreaded, boolean isRequestHandler, boolean isRequestHandlerMultithreaded, boolean isRequestHandlerInvokeOnMessageReceived, IHttpHandler serverHandler, boolean isCloseOnSendingError, HttpClient httpClient) throws IOException {
		super(connection);
		
		setPersistent(true);

		this.handler = serverHandler;
		
		if (isConnectHandler) {
			connectHandlerAdapter = new ConnectHandlerAdapter(isConnectHandlerMultithreaded);
		}

		if (isDisconnectHandler) {
			disconnectHandlerAdapter = new DisconnectHandlerAdapter(isDisconnectHandlerMultithreaded);
		}


		if (isRequestTimeoutHandler) {
			requestTimeoutHandlerAdapter = new RequestTimeoutHandlerAdapter(isRequestTimeoutHandlerMultithreaded);
		}

		this.isRequestHandler = isRequestHandler;
		this.isRequestHandlerMultithreaded = isRequestHandlerMultithreaded;
		this.isRequestHandlerInvokeOnMessageReceived = isRequestHandlerInvokeOnMessageReceived;

		if (httpClient == null) {
			this.httpClient = getDefaultHttpClient();
		} else {
			this.httpClient = httpClient;
		}

		init();
		onConnect();
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId() + "] http server connection established");
		}
	}
	
	
	
	private static synchronized HttpClient getDefaultHttpClient() {
		if (defaultHttpClient == null) {
			defaultHttpClient = new HttpClient();
		}
		
		return defaultHttpClient;
	}
	
	
	public boolean isAutohandleExpect100ContinueHeader() {
		return isAutohandleExpect100ContinueHeader;
	}
	
	public void setAutohandleExpect100ContinueHeader(boolean isAutohandleExpect100ContinueHeader) {
		this.isAutohandleExpect100ContinueHeader = isAutohandleExpect100ContinueHeader;
	}
	
	
	public Executor getWorkerPool() {
		return super.getWorkerpool();
	}


	private void checkRequestTimeout(long currentTimeMillis) throws SocketTimeoutException {
	
		if (currentTimeMillis > (lastTimeMessageReceivedMillis + (((long) requestTimeoutSec) * 1000)) && 
			currentTimeMillis > (getLastTimeWritten() + (((long) requestTimeoutSec) * 1000))) {
			
			onRequestTimeout();
		}
	}

	
	private void onRequestTimeout() {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("request timeout " + requestTimeoutSec + " sec reached");
		}
		terminateWatchDogTask();
		callRequestTimeoutHandler();
	}
	

	/**
	 * set the max transactions per connection. Setting this filed causes that 
	 * a keep-alive response header will be added 
	 * 
	 * @param maxTransactions  the max transactions 
	 */
	public void setMaxTransactions(int maxTransactions) {
		this.maxTransactions = maxTransactions;
	}
	
	
	/**
	 * set the max receive timeout which is accepted for the connection. Setting this
	 * field cases, that a keep-alive response header will be added 
	 * 
	 * @param requestTimout the timeout
	 */
	public void setRequestTimeoutSec(int requestTimeoutSec) {

		if (requestTimeoutSec <= 0) {
			callRequestTimeoutHandler();
			return;
		}
		
		lastTimeMessageReceivedMillis = System.currentTimeMillis();
		
		if ((this.requestTimeoutSec == null) || (this.requestTimeoutSec != requestTimeoutSec)) {
			this.requestTimeoutSec = requestTimeoutSec;
			
			long watchdogPeriod = 0; 
			watchdogPeriod = requestTimeoutSec * 200;
			
			if (watchdogPeriod > MIN_WATCHDOG_PERIOD_MILLIS) {
				watchdogPeriod = MIN_WATCHDOG_PERIOD_MILLIS;
			}
			
			updateWatchDog(watchdogPeriod);
		}
	}


	private synchronized void updateWatchDog(long watchDogPeriod) {
		terminateWatchDogTask();

	    watchDogTask = new TimeoutWatchDogTask(this); 
	    AbstractHttpConnection.schedule(watchDogTask, watchDogPeriod, watchDogPeriod);
	}


	private synchronized void terminateWatchDogTask() {
		if (watchDogTask != null) {
	        watchDogTask.cancel();
	        watchDogTask = null;
	    }
	}

	
	private void callRequestTimeoutHandler() {
		if (requestTimeoutHandlerAdapter != null) {
			requestTimeoutHandlerAdapter.callOnRequestTimeout();
		} else {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getId() + "] receive timeout for http server connection reached. terminate connection (no timeout handler has been set)");
			}

			closeSilence();
		}	
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void close() throws IOException {
		terminateWatchDogTask();
		super.close();
	};
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void destroy() {
		terminateWatchDogTask();
		super.destroy();
	}
	
	
	/**
	 * returns the message receive timeout
	 * 
	 * @return the message receive timeout
	 */
	public int getRequestTimeoutSec() {
		return requestTimeoutSec;
	}	
	

	protected void onMessage(INonBlockingConnection connection) throws BufferUnderflowException, IOException {
		
		assert (Thread.currentThread().getName().startsWith("xDispatcher"));
	
		// reading the header
		HttpRequestHeader requestHeader = HttpRequestHeader.readFrom(connection, MAX_HEADER_LENGTH);
		final HttpRequest request = new HttpRequest(requestHeader);

		
		// handle connection header
		handleConnectionHeaders(request);

	
		// add body parser
		//addFullMessageBodyParser(request, connection);
		addBodyParserIfRequired(request, connection);

		
		countReceivedMessages.incrementAndGet();		
		if (LOG.isLoggable(Level.FINE)) {
			
			if (request.getNonBlockingBody() == null) {
				LOG.fine("[" + connection.getId() + "] bodyless request received from " + connection.getRemoteAddress() +
						 ":" + connection.getRemotePort() + 
						 " (" + countReceivedMessages + ". request) " + requestHeader.toString());
			
			} else {
				String body = "";
				
				String contentType = request.getContentType();
				if (contentType != null) {
					if (contentType.startsWith("application/x-www-form-urlencode")) {
						body = request.getNonBlockingBody().toString() + "\n";
					}
				} 
				
				LOG.fine("[" + connection.getId() + "] request received  from " + connection.getRemoteAddress() + 
						 ":" + connection.getRemotePort() + 
						 " (" + countReceivedMessages + ". request) " + requestHeader.toString() + body);
			}
		}
		
	
		
			
		// handle 100 continue header
		if (isAutohandleExpect100ContinueHeader) {
			if (request.containsHeaderValue("Expect", "100-Continue")) {
				sendBodyless(new HttpResponseHeader(100));
				return;
			}
		}
		 
		
		if (isRequestHandler) {
			if (isContentTypeFormUrlencoded(request)) {
				
				IBodyCompleteListener cl = new IBodyCompleteListener() {
				
					@Execution(Execution.NONTHREADED)
					public void onComplete() throws IOException {
						handleMessage(newFormEncodedRequestWrapper(request));
					}
				};
				request.getNonBlockingBody().addCompleteListener(cl);
				
			} else {
				handleMessage(request);
			}
		}		
	}
	
	
	
	
	private void handleMessage(IHttpRequest request) throws IOException {

		openRequests.incrementAndGet();
		synchronized (requestQueue) {
			requestQueue.add(request);
			requestQueueVersion++;
		}
		
		// does request has a body? 
		if (request.hasBody()) {
			
			request.getNonBlockingBody().addCompleteListener(bodyCompleteListener);

			if (!isRequestHandlerInvokeOnMessageReceived) {
				lastTimeMessageReceivedMillis = System.currentTimeMillis();
				performOnRequestCallBack();
			}
			
		} else {
			suspendRead();
			
			lastTimeMessageReceivedMillis = System.currentTimeMillis();
			performOnRequestCallBack();
		}			
	}
	
	
	
	private boolean isContentTypeFormUrlencoded(IHttpRequest request) {
		
		if (!request.hasBody()) {
			return false;
		}
		
		String contentType = request.getContentType();
		if ((contentType != null) && (contentType.startsWith("application/x-www-form-urlencoded"))) {
			return true;
		}
		
		return false;
	}
	
	
	private void addBodyParserIfRequired(HttpRequest request, INonBlockingConnection connection) throws IOException {
		
		try {
			// body less request by definition
			String method = request.getMethod();
			if (method.equalsIgnoreCase("GET") || method.equalsIgnoreCase("HEAD") || method.equalsIgnoreCase("TRACE")) {
				return;
			}
			
			// contains a non-zero Content-Length header  -> bound body 
			if ((request.getContentLength() != -1)) {
				if (request.getContentLength() > 0) {
					addBoundBodyParser(request);
				}
				return;
			}
			
			//  transfer encoding header is set with chunked -> chunked body
			String transferEncoding = request.getTransferEncoding();
			if ((transferEncoding != null) && (transferEncoding.equalsIgnoreCase("chunked"))) {
				addChunkedBodyParser(request);
				return;
			}
		} catch (BufferUnderflowException ignore) { }
	}
	
	
	private void performOnRequestCallBack() {
	
		if (isRequestHandlerMultithreaded) {
			processMultiThreaded(onRequestCaller);

		} else {
			processNonThreaded(onRequestCaller);
		}
	}
	
	
	void processNonthreaded(Runnable task) {
		super.processNonThreaded(task);
	}
	
	void processMultithreaded(Runnable task) {
		super.processMultiThreaded(task);
	}
	

	@Override
	protected void onBodyDataReceived() {
		lastTimeMessageReceivedMillis = System.currentTimeMillis();
	}




	private void handleConnectionHeaders(HttpRequest request) {


		String keepAliveHeader = request.getHeader("Keep-Alive");
		if (keepAliveHeader != null) {
			String[] tokens = keepAliveHeader.split(",");
			for (String token : tokens) {
				handleKeepAlive(token);
			}
		}


		String connectionHeader = request.getHeader("Connection");
		if (connectionHeader != null) {
			String[] values = connectionHeader.split(",");


			for (String value : values) {
				value = value.trim();

				if (value.equalsIgnoreCase("close")) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("received connection: closed header. destroying connection");
					}
					isCloseAfterResponse = true;
				}
			}
		}
	}


	

	private void handleKeepAlive(String option) {
		try {
			Integer timeoutSec = null; 
		
			if (option.toUpperCase().startsWith("TIMEOUT=")) {
				timeoutSec = Integer.parseInt(option.substring("TIMEOUT=".length(), option.length()));

			} else {
				timeoutSec = Integer.parseInt(option);
			}
			

			if (timeoutSec != null) {
				if (requestTimeoutSec == null) {
					setRequestTimeoutSec(timeoutSec);
				}
			}

		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by handling keep alive option " + option + " " + e.toString());
			}
		}
	}


	/**
	 * send the response. If the content-length header is not set, the message will send 
	 * in chunked mode 
	 *
	 * @param header   the header
	 * @return the body handle to write
	 * @throws IOException if an exception occurs
	 */
	final BodyDataSink send(IHttpResponseHeader header) throws IOException {
		
		if (header.getContentLength() < 0) {
			if (header.getTransferEncoding() == null) {
				header.setTransferEncoding("chunked");
			}
		}

		if (header.getStatus() >= 200) {
			openRequests.decrementAndGet();
		}

		enhanceHeader(header);

		BodyDataSink bodyDataSink = null;
		if (header.getContentLength() < 0) {
			bodyDataSink = newChunkedBodyDataSink(header);
		} else {
			bodyDataSink = newBoundBodyDataSink(header);
		}
		
		 
		bodyDataSink.addCloseListener(new BodyCloseListener(isCloseAfterResponse));
	
		return bodyDataSink;
	}



	/**
	 * send the response in a plain body non-blocking mode
	 *
	 * @param header         the header
	 * @param contentLength  the body content length
	 * @return the body handle to write
	 * @throws IOException if an exception occurs
	 */
	final BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException {
		
		try {
			if ((header.getTransferEncoding() != null) && (header.getTransferEncoding().equalsIgnoreCase("chunked"))) {
				LOG.warning("message header contains Transfer-Encoding: chunked. removing it because message should be sent with predefined length (Content-Length)");
				header.removeHeader("Transfer-Encoding");
			}
	
			
			header.setContentLength(contentLength);
	
			if (header.getStatus() >= 200) {
				openRequests.decrementAndGet();
			}
	
			enhanceHeader(header);
	
			BodyDataSink bodyDataSink = newBoundBodyDataSink(header);
			bodyDataSink.addCloseListener(new BodyCloseListener(isCloseAfterResponse));
	
			return bodyDataSink;
		} catch (IOException ioe) {
			resumeRead();
			throw ioe;
		}
	}

	

	void sendBodyless(IHttpResponseHeader header) throws IOException {
		try {
			if (header.getStatus() >= 200) {
				openRequests.decrementAndGet();
			}
	
			
			enhanceHeader(header);
			
	
			resumeRead();
			
			FlushMode savedFlushmode = getFlushmode();
			setFlushmode(FlushMode.ASYNC);

			writeHeader(header);
			flush();
			
			setFlushmode(savedFlushmode);
			countSendMessages.incrementAndGet();
			
			if (isCloseAfterResponse) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] closing http server connection (isCloseAfterResponse=true");
				}
				close();
			}
		} catch (IOException ioe) {
			resumeRead();
			throw ioe;
		}
	}




	String generateErrorMessageHtml(int errorCode) throws IOException {
		return generateErrorMessageHtml(errorCode, AbstractHttpConnection.getReason(errorCode));
	}




	String generateErrorMessageHtml(int errorCode, String msg) throws IOException {
		msg = msg.replace("\r", "<br/>");

		
		isCloseAfterResponse = true;

		String txt = "<html>\r\n" +
		             "  <!-- This page is auto-generated by xSocket-http (http://xsocket.org) -->\r\n" +
					 "  <head>\r\n" +
					 "    <title>Error " + errorCode + "</title>\r\n" +
					 "    <meta http-equiv=\"cache-control\" content=\"no-cache\"/>" +
					 "  </head>\r\n\r\n" +
					 "  <body>\r\n" +
					 "    <H1 style=\"color:#0a328c;font-size:1.5em;\">ERROR " + errorCode + "</H1>\r\n" +
					 "    <p style=\"font-size:1.5em;\">" + msg + "</p>\r\n" +
					 "    <p style=\"font-size:0.8em;\">" + DATE_FORMAT.format(new Date()) + "    xSocket-http (" + HttpUtils.getVersionInfo() + ")</p>\r\n" +
					 "  <body>\r\n" +
					 "</html>\r\n";
			
		return txt;
	}

	
	@Override
	protected void onProtocolException(Throwable throwable) {
		onRequestHandlingException(throwable);
	}
	

	void onRequestHandlingException(Throwable throwable) {
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("exception occured " + DataConverter.toString(throwable));
		}
		
		isCloseAfterResponse = true;
		
		if (openRequests.get() > 0) {

			try {
				String errorPage = null;
				if (showDetailedError) {
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					throwable.printStackTrace(new PrintStream(baos));
					
					String stackTrace = baos.toString();
					errorPage = generateErrorMessageHtml(500, stackTrace);
	
				} else {
					errorPage = generateErrorMessageHtml(500);
				}
				
				
				BodyDataSink outChannel = send(new HttpResponseHeader(500, "text/html; charset=" + DEFAULT_BODY_ENCODING), errorPage.length());
				outChannel.setFlushmode(FlushMode.ASYNC);
				outChannel.write(errorPage);
				outChannel.close();

			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("could not send error message for " + throwable.toString() + " reason: " + ioe.toString());
				}
			}
		}
	}

	

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {		
		return "HttpServerConnection (requests=" + countSendMessages + "; responses=" + countReceivedMessages + ") " + super.toString();
	}




	private void enhanceHeader(IHttpResponseHeader header) {

		String server = header.getHeader("SERVER");
		if (server == null) {
			header.setHeader("Server", ServerUtils.getComponentInfo());
		}

		int remainingTransactions = 0; 
		if (maxTransactions != null) {
			remainingTransactions = maxTransactions - countReceivedMessages.get();
			if (remainingTransactions <= 0) {
				isCloseAfterResponse = true;
			}
		}

		
		if (!isCloseAfterResponse && (header.getHeader("Connection") == null)) {
			if ((maxTransactions != null) || (requestTimeoutSec != null)) {
				header.setHeader("Connection", "Keep-Alive");
				
				String keepAliveValue = null;
				if (maxTransactions != null) {
					keepAliveValue = "max=" + remainingTransactions;
				}
				
				if (requestTimeoutSec != null) {
					if (keepAliveValue == null) {
						keepAliveValue = "timeout=" + requestTimeoutSec; 
					} else {
						keepAliveValue += ", timeout=" + requestTimeoutSec;
					}
				}
				
				header.setHeader("Keep-Alive", keepAliveValue);
			}
		}

	
		if (header.containsHeaderValue("connection", "close")) {
			isCloseAfterResponse = true;
		} else {
			if (isCloseAfterResponse) {
				header.setHeader("Connection", "close");
			}
			
		}
	}

	
	void sendMsgBody(BodyDataSink bodyDataSink, NonBlockingBodyDataSource bodyDataSource) throws IOException {
		sendMessageBody(bodyDataSink, bodyDataSource);
	}
	
	
	void resumeReadOp() throws IOException {
		suspendRead();
	}


	private void onConnect() throws IOException {
		if (connectHandlerAdapter != null) {
			connectHandlerAdapter.callOnConnect();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onDisconnect() throws IOException {
		destroy();
		
		if (disconnectHandlerAdapter != null) {
			disconnectHandlerAdapter.callOnDisconnect();
		}
		
		super.onDisconnect();
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId() + "] http server connection destroyed");
		}
	}

	

	protected static void schedule(TimerTask task, long delay, long period) {
		AbstractHttpConnection.schedule(task, delay, period);
	}

	final class ConnectHandlerAdapter implements Runnable {

		private boolean isMultithreaded = true;

		public ConnectHandlerAdapter(boolean isMultithreaded) {
			this.isMultithreaded = isMultithreaded;
		}

		private void callOnConnect() {

			if (isMultithreaded) {
				processMultiThreaded(this);
			} else {
				processNonThreaded(this);
			}
		}

		public void run() {
			try {
				((IHttpConnectHandler) handler).onConnect(HttpServerConnection.this);
			} catch (Exception e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] error occured by calling on connect " + handler + " " + e.toString());
				}
			}
		}
	}


	final class DisconnectHandlerAdapter implements Runnable {

		private boolean isMultithreaded = true;

		public DisconnectHandlerAdapter(boolean isMultithreaded) {
			this.isMultithreaded = isMultithreaded;
		}

		private void callOnDisconnect() {

			if (isMultithreaded) {
				processMultiThreaded(this);
			} else {
				processNonThreaded(this);
			}
		}

		public void run() {
			try {
				((IHttpDisconnectHandler) handler).onDisconnect(HttpServerConnection.this);
			} catch (Exception e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] error occured by calling on connect " + handler + " " + e.toString());
				}
			}
		}
	}


	final class RequestTimeoutHandlerAdapter implements Runnable {

		private boolean isMultithreaded = true;

		public RequestTimeoutHandlerAdapter(boolean isMultithreaded) {
			this.isMultithreaded = isMultithreaded;
		}

		private void callOnRequestTimeout() {

			if (isMultithreaded) {
				processMultiThreaded(this);
			} else {
				processNonThreaded(this);
			}
		}

		public void run() {
			try {
				boolean isHandled = ((IHttpRequestTimeoutHandler) handler).onRequestTimeout(HttpServerConnection.this);
				if (!isHandled) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("receive timeout for http server connection reached. terminate connection (timeout handler returned false)");
					}

					closeSilence();
				}
			} catch (Exception e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] error occured by calling on request timeout " + handler + " " + e.toString());
				}
			}
		}
	}
	
	
	private final class BodyCompleteListener implements IBodyCompleteListener {
		
		@Execution(Execution.NONTHREADED)
		public final void onComplete() {
			lastTimeMessageReceivedMillis = System.currentTimeMillis();
			
			if (countSendMessages.get() < countReceivedMessages.get()) {
				try {
					suspendRead();
				} catch (IOException ioe) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("[" + getId() + "] error occured by suspending read " + DataConverter.toString(ioe)); 
					}
				}
			}

			// call handlers call back method if is HEADER_MESSAGE triggered
			if (isRequestHandlerInvokeOnMessageReceived)  {
				performOnRequestCallBack();
			}
		}
	}

	
	private final class BodyCloseListener implements IBodyCloseListener {
		
		private boolean closeConnection = false;
		
		public BodyCloseListener(boolean closeConnection) {
			this.closeConnection = closeConnection;
		}
		
		@Execution(Execution.NONTHREADED)
		public void onClose() throws IOException {
			
			countSendMessages.incrementAndGet();
			
			if (closeConnection) {
				closeSilence();
			} else {
				resumeRead();
			}
		}
	}


	private final class OnRequestCaller implements Runnable {
		
		
		public void run() {
			
			IHttpRequest request = null;
			int currentRequestQueueVersion = 0;

			do {
				
				// get first message of the queue 
				synchronized (requestQueue) {
					
					// are messages in the queue?
					if (requestQueue.isEmpty()) {
						
						// .. no, do nothing 
						return;
						
					// ... yes, messages are in the queue
					} else {
						// get (but do not remove) message from the queue
						request = requestQueue.get(0);
						
						// is message has not been complete received and is mode MESSAGE do nothing 
						if (isRequestHandlerInvokeOnMessageReceived && !isMessageComplete(request)) {
							return;
						}
						
						// remove the message from the queue and update counters 
						requestQueue.remove(0);
						requestQueueVersion++;
						currentRequestQueueVersion = requestQueueVersion;
					} 
				}


				// perform call back method
				try {
					((IHttpRequestHandler) handler).onRequest(new Exchange(HttpServerConnection.this, request, httpClient));
				} catch (Exception e) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("[" +getId() + "] error occured by calling on request " + handler + " " + e.toString());
					}
					
					onRequestHandlingException(e);
				}


				// if meantime a new request is received -> loop
				synchronized (requestQueue) {
					if (currentRequestQueueVersion == requestQueueVersion) {
						break;
					}
				}

			} while (request != null);			
		}
	}

	
	private boolean isMessageComplete(IHttpRequest request) {
		
		// does the message contain a body?
		if (request.hasBody()) {
			try {
				return isComplete(request.getNonBlockingBody());
			} catch (IOException ioe) {
				return false;
			}
			
        // ... no it is a bodyless message			
		} else {
			return true;
		}
	}

	
	
	
	private static final class TimeoutWatchDogTask extends TimerTask {
		
		private WeakReference<HttpServerConnection>  connectionRef = null;
		
		public TimeoutWatchDogTask(HttpServerConnection connection) {
			connectionRef = new WeakReference<HttpServerConnection>(connection);
		}
	
		
		@Override
		public void run() {
			try {
				HttpServerConnection connection = connectionRef.get();
				
				if (connection == null)  {
					this.cancel();
					
				} else {
					connection.checkRequestTimeout(System.currentTimeMillis());
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}		
	}
	
	
	

	/**
	 *
	 *
	 * @author grro@xsocket.org
	 */
	private static final class Exchange implements IHttpExchange {

		private Logger LOG = Logger.getLogger(Exchange.class.getName());
		
		
		private HttpServerConnection httpConnection = null;
		private IHttpRequest httpRequest = null;
		
		private HttpClient httpClient = null;
		
		private boolean isResponseCommitted = false;

		Exchange(HttpServerConnection httpConnection, IHttpRequest httpRequest, HttpClient httpClient) {
			this.httpConnection = httpConnection;
			this.httpRequest = httpRequest;
			this.httpClient = httpClient;
		}
		
		
		
		public IHttpRequest getRequest() {
			return httpRequest;
		}
		
		
		/**
		 * send the response. If the content-length header is not set, the message will send 
		 * in chunked mode 
		 *
		 * @param header   the header
		 * @return the body handle to write
		 * @throws IOException if an exception occurs
		 */
		public BodyDataSink send(IHttpResponseHeader header) throws IOException {
			isResponseCommitted = true;
			BodyDataSink bodyDataSink = httpConnection.send(header);
			
			return bodyDataSink;
		}
		
		

		

		/**
		 * send the response in a plain body non-blocking mode
		 *
		 * @param header         the header
		 * @param contentLength  the body content length
		 * @return the body handle to write
		 * @throws IOException if an exception occurs
		 */
		public BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException {
			isResponseCommitted = true;
			BodyDataSink bodyDataSink = httpConnection.send(header, contentLength);	

			return bodyDataSink;
		}

		

		/**
		 * send the response
		 *
		 * @param response   the response
		 * @throws IOException if an exception occurs
		 */
		public void send(IHttpResponse response) throws IOException {
			
			isResponseCommitted = true;

			
			try {
				
				IHttpResponseHeader responseHeader = response.getResponseHeader();
		
		
				// body less request?
				if (response.getNonBlockingBody() == null) {
					sendResponseWithoutBody(response);
		
				// no, request has a body
				} else {
					
					if (response.getNonBlockingBody().getDataHandler() != null) {
						LOG.warning("a body handler is already assigned to the message body. current body handler will be removed");
					}
		
					
					// is old style (pre Http/1.0) header-less response?    
					if (responseHeader.getStatus() == 0) {
						responseHeader  = new HttpResponseHeader(200);   // creates a HTTP/1.1 header
						response = new HttpResponse(responseHeader, response.getNonBlockingBody());   

					// is Http/1.0 response and connection terminated?	
					} if (!response.getProtocol().equalsIgnoreCase("HTTP/1.1") && (response.getContentLength()== -1)) {	
						response.setProtocol("HTTP/1.1");   // response will be send in chunked mode
					} 
				
					sendResponseWithBody(response);				
				}
			} catch (IOException ioe) {
				httpConnection.resumeReadOp();
				throw ioe;
			}
		}
		
		
		private void sendResponseWithoutBody(IHttpResponse response) throws IOException {		
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + httpConnection.getId() + "] sending (bodyless): " + response.getResponseHeader());
			}

			if (response.getContentLength() == -1) {
				response.setContentLength(0);
			}

			
		
			httpConnection.sendBodyless(response.getResponseHeader());
		}
		
		
		private void sendResponseWithBody(IHttpResponse response) throws IOException {
			BodyDataSink bodyDataSink = null;
			if (response.getContentLength() >= 0) {									
				bodyDataSink = send(response.getResponseHeader(), response.getContentLength());
			} else {
				bodyDataSink = send(response.getResponseHeader());
			}
			
			


			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + httpConnection.getId() + "] sending (" + bodyDataSink.getClass().getSimpleName() +
						 "): " + response.getResponseHeader());
			}

			httpConnection.sendMsgBody(bodyDataSink, response.getNonBlockingBody());
		}


		/**
		 * send the response
		 *
		 * @param status   the status
		 * @throws IOException if an exception occurs
		 */
		void send(int status) throws IOException {
			send(new HttpResponse(status));
		}
		

		

		/**
		 * send an error response
		 *
		 * @param errorCode   the error code
		 */
		public void sendError(int errorCode) {
			try {
				send(new HttpResponse(errorCode, "text/html; charset=" + HttpServerConnection.DEFAULT_BODY_ENCODING, httpConnection.generateErrorMessageHtml(errorCode)));
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("could not send error message " + errorCode + " reason " + ioe.toString());
				}
			}
		}

		

		public BodyDataSink forward(IHttpRequestHeader requestHeader, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			return httpClient.send(requestHeader, responseHandler);
		}


		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			return httpClient.send(requestHeader, contentLength, responseHandler);
		}


		public void forward(IHttpRequest request, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			httpClient.send(request, responseHandler);		
		}

		
		public void destroy() {
			httpConnection.destroy();
		}

	

		/**
		 * send an error response
		 *
		 * @param errorCode   the error code
		 * @param msg         the error message
		 */
		public void sendError(int errorCode, String msg) {
			try {
				send(new HttpResponse(errorCode, "text/html; charset=" + HttpServerConnection.DEFAULT_BODY_ENCODING, httpConnection.generateErrorMessageHtml(errorCode, msg)));
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("could not send error message " + errorCode + " reason " + ioe.toString());
				}
			}

		}


		/**
		 * returns the underlying connection
		 *
		 * @return the connection
		 */
		public HttpServerConnection getConnection() {
			return httpConnection;
		}
		
		
		void processNonhreaded(Runnable task) {
			httpConnection.processNonthreaded(task);
		}
		
		void processMultithreaded(Runnable task) {
			httpConnection.processMultithreaded(task);
		}
		

		boolean isResponseCommitted() {
			return isResponseCommitted;
		}
		
		void onRequestHandlingException(Throwable throwable) {
			httpConnection.onRequestHandlingException(throwable);
		}
	}
	 
}
